#!/usr/bin/env python3

__author__ = "Samuel Murail"


from gromacs_py import gmx
import argparse


def parser_input():

    # Parse arguments :
    parser = argparse.ArgumentParser(description="Simulation production")
    parser.add_argument('-f', action="store", dest="f",
                        help='Input PDB file', type=str,
                        required=True)
    parser.add_argument('-p', action="store", dest="p",
                        help='Topologie in gromacs format .top',
                        type=str, required=True)
    parser.add_argument('-o', action="store", dest="o",
                        help='Output Directory',
                        type=str,
                        required=True)
    parser.add_argument('-n', action="store", dest="name",
                        help='Output file name',
                        type=str,
                        required=True)
    parser.add_argument('-time', action="store", dest="time",
                        help='Production time, default=10',
                        type=float, default=10)
    parser.add_argument('-dt', action="store", dest="dt",
                        help='Equilibration  dt, default=0.002 (2 fs)',
                        type=float, default=0.002)
    parser.add_argument('-maxwarn', action="store", dest="maxwarn",
                        help='Total number of warnings allowed for the '
                        'equilibration,'
                        ' default=0',
                        type=int, default=0)
    parser.add_argument('-nt', action="store", dest="nt",
                        help='Total number of threads to start, default=0',
                        type=float, default=0)
    parser.add_argument('-ntmpi', action="store", dest="ntmpi",
                        help='Number of thread-MPI threads to start, '
                        'default=0',
                        type=float, default=0)
    parser.add_argument('-gpu_id', action="store", dest="gpuid",
                        help='List of GPU device id-s to use, default=\"\" ',
                        default="None")
    return(parser)


if __name__ == "__main__":

    parser = parser_input()
    args = parser.parse_args()

    dt = args.dt
    nsteps = 1000 * args.time / dt
    maxwarn = args.maxwarn

    sys_prod = gmx.GmxSys(name=args.name, coor_file=args.f, top_file=args.p)
    sys_prod.nt = args.nt
    sys_prod.ntmpi = args.ntmpi
    if args.gpuid != "None":
        sys_prod.gpu_id = args.gpuid

    sys_prod.production(out_folder=args.o, name=args.name, nsteps=nsteps,
                        dt=args.dt, maxwarn=maxwarn)

    print("\n\nProductuion was sucessfull \n\tProduction directory :\t"
          + args.o)

    sys_prod.display()



#!/usr/bin/env python3
# coding: utf-8
##################################
# #######   GROMACS 5   ##########
##################################


import os
import logging
import sys

from shutil import copy as shutil_copy

from os_command_py import os_command

from .itp import Itp

# Autorship information
__author__ = "Samuel Murail"
__copyright__ = "Copyright 2020, RPBS"
__credits__ = ["Samuel Murail"]
__license__ = "GNU General Public License v2.0"
__maintainer__ = "Samuel Murail"
__email__ = "samuel.murail@u-paris.fr"
__status__ = "Production"

# Logging
logger = logging.getLogger(__name__)


# Check if Readthedoc is launched skip the program path searching
on_rtd = os.environ.get('READTHEDOCS') == 'True'
if on_rtd:
    logger.info("Gromacs cannot be found")
    GMX_BIN = ""
    gmx_version = ""
else:
    GMX_BIN = os_command.which('gmx')
    gmx_version = os_command.get_gmx_version()
    logger.info("Gromacs version is {}".format(gmx_version))

# Get local gmx path
GMX_PATH = "/".join(GMX_BIN.split("/")[:-2])

# Get library path
GROMACS_MOD_DIRNAME = os.path.dirname(os.path.abspath(__file__))

FORCEFIELD_PATH_LIST = [os.path.join(GROMACS_MOD_DIRNAME, "template")]
# GMXLIB_var should not include GMXPATH top dir, it could induce some conflict
GMXLIB_var = os.path.join(GROMACS_MOD_DIRNAME, "template")

# Check if GMXLIB env variable is defines if yes add it to forcefield path
if 'GMXLIB' in os.environ:
    # Needed for pytest in monitor.py, otherwise add twice GROMACS_MOD_DIRNAME
    if os.environ['GMXLIB'] not in FORCEFIELD_PATH_LIST:
        FORCEFIELD_PATH_LIST.append(os.environ['GMXLIB'])
        GMXLIB_var += ":" + os.environ['GMXLIB']

FORCEFIELD_PATH_LIST.append(os.path.join(GMX_PATH, "share/gromacs/top"))


def show_log():
    """ To use only with Doctest !!!
    Redirect logger output to sys.stdout
    """
    # Delete all handlers
    logger.handlers = []
    # Set the logger level to INFO
    logger.setLevel(logging.INFO)
    # Add sys.sdout as handler
    logger.addHandler(logging.StreamHandler(sys.stdout))

    # Show log of top sys:
    from . import itp
    itp.show_log()


def show_debug(pdb_manip_log=True):
    """ To use only with Doctest !!!
    Redirect logger output to sys.stdout
    """
    # Delete all handlers
    logger.handlers = []
    # Set the logger level to INFO
    logger.setLevel(logging.DEBUG)
    # Add sys.sdout as handler
    logger.addHandler(logging.StreamHandler(sys.stdout))


class TopSys:
    """Topologie base on gromacs .top :
    #include forcefield

    All name and full path of itp are save
    [ system ] -> Name
    [ molecules ] -> Composition

    :param path: topologie file path
    :type path: str

    :param forcefield: name of the focefield
    :type forcefield: str

    :param itp_list: list of the itp object
    :type itp_list: list

    :param mol_comp: molecular composition
    :type mol_comp: list

    :param name: name of the system
    :type tpr: str

    :param folder: path of the top file folder
    :type folder: str

    :param include_itp: Flag indicating if the topologie include a molecule
        topologie
    :type include_itp: bool
    """

    def __init__(self, top_in):

        self.path = os_command.full_path_and_check(top_in)
        self.forcefield = None
        self.itp_list = []
        self.mol_comp = []
        self.name = ""
        self.folder = os_command.get_directory(top_in)
        self.include_itp = False
        # Intermolecular restraints (only for Free ener):
        self.inter_interact = False
        self.inter_bond_list = []
        self.inter_angl_list = []
        self.inter_dihe_list = []

        self.read_file(top_in)
        if self.include_itp:
            logger.info("Rewrite topologie: {}".format(top_in))
            self.write_file(top_in)

    def read_file(self, top_in):

        field = None
        ifdef = False
        bond_list = []
        angle_list = []
        dihed_list = []

        with open(top_in) as topfile:
            for line in topfile:
                # print("line: ",line)
                # Check #ifdef field:
                if line[:6] == '#ifdef':
                    ifdef = True
                if line[:6] == '#endif':
                    ifdef = False
                # Check include files
                if not ifdef and line.startswith('#include'):
                    # get the file name to include:
                    file_name = line.split()[1][1:-1]
                    # remove '"' , path and .itp:
                    include = (file_name.split("/")[-1]).split(".")[0]

                    # Look if the itp is in the top path:
                    # print("Path 1: ",self.folder+"/"+file_name)
                    # print("Path 2: ",FORCEFIELD_PATH_LIST+"/"+file_name)

                    itp_found = False
                    if os_command.check_file_exist(
                            os.path.join(self.folder, file_name)):
                        path = os.path.abspath(os.path.join(self.folder,
                                                            file_name))
                        itp_found = True
                    else:
                        for forcefield in FORCEFIELD_PATH_LIST:
                            if os_command.check_file_exist(
                                    os.path.join(forcefield, file_name)):
                                path = os.path.abspath(
                                    os.path.join(forcefield, file_name))
                                itp_found = True
                                break
                    if not itp_found:
                        raise IOError('Itp ' + file_name + ' not found')

                    # print("name =", include, "fullname =", file_name,
                    # "path =",path)
                    if include == "forcefield":
                        self.forcefield = {'name': file_name.split('.')[0],
                                           'fullname': file_name,
                                           'path': path}
                    else:
                        self.add_mol_itp(path)
                # Check field
                elif not ifdef and line.startswith("["):
                    # Remove space and [ ]
                    field = line.strip()[1:-1].strip()
                    # print(field)
                    # As intermolecular_inter field is empty
                    # (filled with bonds, angles, ...)
                    # It has to be checked now:
                    if field == 'intermolecular_interactions':
                        self.inter_interact = True
                    continue
                # Check in the field :
                elif (not ifdef and not line[0].startswith(";") and
                      line.strip() != ""):
                    if field == 'moleculetype':
                        # in the case where mol param are present in the top
                        # file, convert the top to an itp
                        # If mol_name variable is defined, give to the mol
                        # param this name
                        name_itp = os.path.basename(top_in)[:-3] + "itp"

                        logger.info("Molecule topologie present in {} "
                                    ", extract the topologie in a separate"
                                    " file: {}".format(top_in, name_itp))
                        # Store and write the itp file:
                        top_itp = Itp(name=name_itp, fullname=name_itp,
                                      path=os.path.abspath(top_in))
                        top_itp.write_file(os.path.join(self.folder, name_itp))
                        top_itp.display()
                        # Add the itp to the itp_list
                        self.add_mol_itp(os.path.join(self.folder, name_itp))
                        self.include_itp = True
                    # Name of the system
                    elif field == 'system':
                        self.name = line.strip()
                    # Molecule composition of the system
                    elif field == 'molecules':
                        line_list = line.strip().split()
                        self.mol_comp.append({'name': line_list[0],
                                              'num': line_list[1]})
                    # Following command concern
                    # intermolecular interactions
                    elif field == 'bonds':
                        bond_list.append(line.strip().split())
                    elif field == 'angles':
                        angle_list.append(line.strip().split())
                    elif field == 'dihedrals':
                        dihed_list.append(line.strip().split())
        if self.inter_interact:
            self.add_intermolecular_restr(bond_list=bond_list,
                                          angle_list=angle_list,
                                          dihed_list=dihed_list)

    def display(self):
        if self.forcefield:
            print("Forcefield include :\n", self.forcefield['name'])
        else:
            print('No forcefield defined !')
        for itp in self.itp_list:
            itp.display()
        print("Mol List:")
        for mol in self.mol_comp:
            print("   * {} {}".format(mol['num'], mol['name']))

        print("Mol Name:\n", self.name)

    def write_file(self, top_out):
        filout = open(top_out, 'w')

        filout.write("; Topologie file created by " + __name__ + "\n")
        if self.forcefield:
            filout.write("; Forcefield: \n")
            filout.write("#include \"" + self.forcefield['fullname'] + "\"\n")
        # print include files
        filout.write("\n; Itp to include: \n")
        for itp in self.itp_list:
            filout.write("#include \"" + itp.fullname + "\"\n")
            # Check if the include is in the ff folder:
            # if self.forcefield['fullname'].split("/")[0] == \
            #       itp.fullname.split("/")[0]:
            #    filout.write("#include \""+itp.fullname+"\"\n")
            # else:
            #    filout.write("#include \""+itp.fullname+"\"\n")
        filout.write("\n[ system ]\n; Name\n")
        filout.write(self.name + "\n")
        filout.write("\n[ molecules ]\n; Compound        #mols\n")
        for mol in self.mol_comp:
            filout.write(mol['name'] + " \t" + mol['num'] + "\n")
        filout.write("\n")
        # INTER
        if self.inter_interact:
            filout.write("[ intermolecular_interactions ]\n")
            if self.inter_bond_list:
                filout.write("\n[ bonds ]\n;  ai    aj    type   "
                             " bA    kA    bB    kB\n")
                for param in self.inter_bond_list:
                    filout.write(
                        "{:>6}{:>6}{:>6}{:>13}{:>13}{:>13}{:>13}\n".format(
                            param['ai'], param['aj'], param['type'],
                            param['rA'], param['kA'],
                            param['rB'], param['kB']))
            if self.inter_angl_list:
                filout.write("\n[ angles ]\n;  ai    aj    ak  "
                             "  type    thA    kA    thB    kB\n")
                for param in self.inter_angl_list:
                    filout.write("{:>6}{:>6}{:>6}{:>6}{:>13}{:>13}{:>13}"
                                 "{:>13}\n".format(
                                    param['ai'], param['aj'],
                                    param['ak'], param['type'],
                                    param['thA'], param['kA'],
                                    param['thB'], param['kB']))
            if self.inter_dihe_list:
                filout.write("\n[ dihedrals ]\n;  ai    aj    ak    al"
                             "    type    phiA    kA    phiB    kB\n")
                for param in self.inter_dihe_list:
                    filout.write("{:>6}{:>6}{:>6}{:>6}{:>6}{:>13}{:>13}"
                                 "{:>13}{:>13}\n".format(
                                    param['ai'], param['aj'],
                                    param['ak'], param['al'],
                                    param['type'],
                                    param['thA'], param['kA'],
                                    param['thB'], param['kB']))

        filout.close()
        self.copy_dependancies(os_command.get_directory(top_out))

    def charge(self):
        """Get the charge of the system
        """

        self._charge = 0
        # Look for mol name
        for mol in self.mol_comp:
            name = mol['name']
            num = int(mol['num'])
            # print("Mol top info: ",name,num)
            # Search mol name in itp:
            for local_itp in self.itp_list:
                itp_charge = local_itp.charge(name)
                if itp_charge is not None:
                    logger.debug("Get charge of {} : {} total charge:"
                                 " {}".format(name, itp_charge,
                                              itp_charge * num))
                    self._charge += num * itp_charge
                    break
        return self._charge

    def prot_res_num(self, selection="Protein"):
        """Compute the residue number of a selection
        """

        self.res_num = 0
        # Look for mol name
        for mol in self.mol_comp:
            name = mol['name']
            num = int(mol['num'])
            logger.info("{} : {}".format(name, num))
            if name.find(selection) != -1:
                # Search mol name in itp:
                for itp in self.itp_list:
                    # print(itp.name)
                    itp_res_num = itp.res_num(name)
                    if itp_res_num is not None:
                        logger.info("Get Res num of {} : {}\n"
                                    "Total number of residue: {}".format(
                                        name, itp_res_num, num * itp_res_num))
                        self.res_num += num * itp_res_num
                        break
        return self.res_num

    def mol_num(self, name):
        """Get the number of the molecule "name"
        """

        mol_num = 0

        for mol in self.mol_comp:
            if mol['name'] == name:
                mol_num = mol_num + int(mol['num'])
        return mol_num

    def add_posre(self, posre_name="POSRE_CA",
                  selec_dict={'atom_name': ['CA']},
                  fc=[1000, 1000, 1000]):
        """Add position restraint based on the selection for each itp
        """
        for mol in self.mol_comp:
            for itp in self.itp_list:
                itp.add_posre(mol_name=mol['name'], posre_name=posre_name,
                              selec_dict=selec_dict, fc=fc)

    def add_mol_itp(self, mol_itp_file):
        """Add a molecule itp in the topologie itp_list.
        """
        fullname = (mol_itp_file.split("/")[-1])
        include = fullname.split(".")[0]
        path = os_command.full_path_and_check(mol_itp_file)
        mol_itp = Itp(name=include, fullname=fullname, path=path)

        # 1. Get new mol name:
        mol_name_list = []
        for top_mol in mol_itp.top_mol_list:
            mol_name_list.append(top_mol.name)

        # 2. Check if it already present
        present = False
        for itp in self.itp_list:
            for top_mol in itp.top_mol_list:
                if top_mol.name in mol_name_list:
                    present = True
                    break
            if present:
                break

        # Add the itp if not present
        if not present:
            self.itp_list.append(mol_itp)

    def add_mol(self, mol_name, mol_itp_file, mol_num):
        """Add a molecule in the topologie (composition and itp_list)
        """

        logger.info("Add {} mol {}".format(mol_num, mol_itp_file))
        self.mol_comp.append({'name': mol_name, 'num': str(mol_num)})
        self.add_mol_itp(mol_itp_file)

    def remove_ion(self, ion_name_list):
        """Remove a molecule from the topologie (composition and itp_list)
        """

        logger.info("Remove mol(s) : {}".format(' '.join(ion_name_list)))
        itp_remove_list = []
        mol_remove_list = []

        for itp in self.itp_list:
            for mol in itp.top_mol_list:
                for atom in mol.atom_dict.values():
                    if atom['atom_name'] in ion_name_list and (
                            len(itp.top_mol_list) == 1):
                        # print(itp.name, 'YOYOYO')
                        itp_remove_list.append(itp)
                        mol_remove_list.append(mol.name)

        for itp in itp_remove_list:
            self.itp_list.remove(itp)

        mol_comp = []
        for mol in self.mol_comp:
            if mol['name'] not in mol_remove_list:
                mol_comp.append(mol)

        self.mol_comp = mol_comp

    def add_atomtypes(self, new_atomtypes):
        """ Add atomtypes in a topologie.

        :param new_atomtypes: path of the atomtype itp file
        :type new_atomtypes: str

        """

        # check if and atomtypes file exists:
        # self.display()
        atom_type = False
        for itp_file in self.get_include_no_posre_file_list():
            if itp_file.split("/")[-1].endswith('atomtypes.itp'):
                # print('atomtypes file present', itp_file)
                atom_type = True
                atomtype_path = itp_file
                break

        # If it doesn't exist create it
        if not atom_type:
            fullname = (new_atomtypes.split("/")[-1])
            include = fullname.split(".")[0]
            path = os_command.full_path_and_check(new_atomtypes)
            atomtype_itp = Itp(name=include, fullname=fullname, path=path)

            self.itp_list = [atomtype_itp] + self.itp_list
        # If it does exist add the new atomtypes in the first one:
        else:
            # First extract old atom types:
            field = None
            atom_dict = {}
            name_list = []
            with open(atomtype_path) as file:
                for line in file:
                    if line.strip().startswith("["):
                        # Remove space and [ ], remove also comments
                        field = line.replace(" ", "").split("]")[0][1:]
                        continue

                    if (line[0] != ";" and line[0] != "#" and
                            line.strip() != ""):
                        # Remove commentary in the line
                        line_comment = line.split(';')
                        line_list = line_comment[0].split()

                        if field == 'atomtypes':
                            name_list.append(line_list[0])
                            name, bond_type, mass, charge, ptype, sigma, \
                                epsilon = line_list[:7]
                            atom_dict[name] = {'bond_type': bond_type,
                                               'mass': float(mass),
                                               'charge': float(charge),
                                               'ptype': ptype,
                                               'sigma': float(sigma),
                                               'epsilon': float(epsilon)}
                            name_list.append(name)

            # Second extract new atom types and
            # check if they are already present:
            field = None
            new_atom_dict = {}
            with open(new_atomtypes) as file:
                for line in file:
                    if line.strip().startswith("["):
                        # Remove space and [ ], remove also comments
                        field = line.replace(" ", "").split("]")[0][1:]
                        continue

                    if (line[0] != ";" and line[0] != "#" and
                            line.strip() != ""):
                        # Remove commentary in the line
                        line_comment = line.split(';')
                        line_list = line_comment[0].split()

                        if field == 'atomtypes':
                            name, bond_type, mass, charge, ptype, sigma, \
                                epsilon = line_list[:7]
                            local_dict = {'bond_type': bond_type,
                                          'mass': float(mass),
                                          'charge': float(charge),
                                          'ptype': ptype,
                                          'sigma': float(sigma),
                                          'epsilon': float(epsilon)}
                            if name not in name_list:
                                new_atom_dict[name] = local_dict
                            else:
                                # Check if the new values are the same:
                                if local_dict != atom_dict[name]:
                                    logger.warning(
                                        'Atom types parameters for {} are '
                                        'different in {} and {}. Only one '
                                        'version is kept !!!'.format(
                                            name, atomtype_path,
                                            new_atomtypes))

            # Finally append the new param in the old one
            with open(atomtype_path, 'a') as file:
                for name, atom_dict in new_atom_dict.items():
                    file.write(' {:3}      {:3}         {:.5f}  {:.5f}'
                               '   {}     {:.5e}   {:.5e}\n'.format(
                                name,
                                atom_dict['bond_type'],
                                atom_dict['mass'],
                                atom_dict['charge'],
                                atom_dict['ptype'],
                                atom_dict['sigma'],
                                atom_dict['epsilon']))

    def add_intermolecular_restr(self, bond_list=[],
                                 angle_list=[], dihed_list=[]):
        """ Add inter molecular restraints in topologie file
        """

        self.inter_interact = True
        for bond in bond_list:
            ai, aj, funct, rA, kA, rB, kB = bond
            self.inter_bond_list.append({'ai': ai, 'aj': aj,
                                         'type': funct,
                                         'rA': rA, 'kA': kA,
                                         'rB': rB, 'kB': kB})
        for angle in angle_list:
            ai, aj, ak, funct, thA, kA, thB, kB = angle
            self.inter_angl_list.append({'ai': ai, 'aj': aj,
                                         'ak': ak, 'type': funct,
                                         'thA': thA, 'kA': kA,
                                         'thB': thB, 'kB': kB})
        for dihed in dihed_list:
            ai, aj, ak, al, funct, thA, kA, thB, kB = dihed
            self.inter_dihe_list.append({'ai': ai, 'aj': aj,
                                         'ak': ak, 'al': al,
                                         'type': funct,
                                         'thA': thA, 'kA': kA,
                                         'thB': thB, 'kB': kB})

    def change_mol_num(self, mol_name, mol_num):
        """ Update molecule number.
        And remove multiple molecule definition if they are consecutive.
        """

        for i, mol in enumerate(self.mol_comp):

            if mol['name'] == mol_name:
                mol['num'] = str(mol_num)
                # Find other mol_name and remove them:
                to_remove_index_list = []
                for j in range(i + 1, len(self.mol_comp)):
                    if self.mol_comp[j]['name'] == mol_name:
                        to_remove_index_list.append(j)
                    else:
                        break
                if len(to_remove_index_list) > 0:
                    for index in sorted(to_remove_index_list, reverse=True):
                        del self.mol_comp[index]
                break
        logger.info(self.mol_comp)

    def get_include_file_list(self):
        file_list = []
        # print("\n\nIncluded files: ")
        for itp in self.itp_list:
            if self.forcefield:
                if (self.forcefield['fullname'].split("/")[0] !=
                        itp.fullname.split("/")[0]):
                    # print(itp.path)
                    file_list.append(itp.path)
                    file_list = file_list + itp.get_include_file_list()
            else:
                file_list.append(itp.path)
                file_list = file_list + itp.get_include_file_list()

        return file_list

    def get_include_no_posre_file_list(self):
        file_list = []
        # print("\n\nIncluded files: ")
        for itp in self.itp_list:
            if self.forcefield:
                if (self.forcefield['fullname'].split("/")[0] !=
                        itp.fullname.split("/")[0]):
                    file_list.append(itp.path)
            else:
                file_list.append(itp.path)
        return file_list

    def copy_top_and_dependancies(self, dest_file):

        dest_folder = os_command.get_directory(dest_file)
        logger.info("Copy topologie file and dependancies")

        if self.path != os.path.abspath(dest_file):
            shutil_copy(self.path, os.path.abspath(dest_file))

        self.copy_dependancies(dest_folder)

    def copy_dependancies(self, dest_folder):

        dest_folder = os.path.abspath(dest_folder)

        file_to_copy = self.get_include_file_list()
        # print(file_to_copy)
        for file in file_to_copy:
            if os.path.abspath(os_command.get_directory(file)) !=\
                    os.path.abspath(dest_folder):
                # print("Copy: "+file+":to: "+dest_folder)
                shutil_copy(file, dest_folder)

    def change_mol_name(self, old_name, new_name):

        # Change name in mol_comp
        name_change_flag = False
        for mol in self.mol_comp:
            if mol['name'] == old_name:
                name_change_flag = True
                mol['name'] = new_name
        # Change in itp :
        for itp in self.itp_list:
            itp.change_mol_name(old_name, new_name)

        if name_change_flag:
            self.write_file(self.path)



#!/usr/bin/env python3
# coding: utf-8

"""
Location of datafiles for gromacs_py tests

Use it as:

```
from gromacs_py.test.datafiles import *
```
"""

import os

# Autorship information
__author__ = "Samuel Murail"
__copyright__ = "Copyright 2020, RPBS"
__credits__ = ["Samuel Murail"]
__license__ = "GNU General Public License v2.0"
__maintainer__ = "Samuel Murail"
__email__ = "samuel.murail@u-paris.fr"
__status__ = "Production"

PYTEST_DIR = os.path.dirname(os.path.abspath(__file__))
TEST_FILE_PATH = os.path.join(PYTEST_DIR, "../test_files/")

PDB_1D30 = os.path.join(TEST_FILE_PATH, '1D30.pdb')
PDB_1Y0M = os.path.join(TEST_FILE_PATH, '1y0m.pdb')
PDB_1RXZ = os.path.join(TEST_FILE_PATH, '1rxz.pdb')
PDB_1JD4 = os.path.join(TEST_FILE_PATH, '1jd4.pdb')



#!/usr/bin/env python3
# coding: utf-8

""" gmx library include the gromacs system class ``GmxSys``, as well as
topologie ``TopSys`` and itp.
"""

import logging

from .itp import Itp
from .topsys import TopSys
from .gmxsys import GmxSys, GROMACS_MOD_DIRNAME, GMX_BIN, gmx_version

# Autorship information
__author__ = "Samuel Murail"
__copyright__ = "Copyright 2020, RPBS"
__credits__ = ["Samuel Murail"]
__license__ = "GNU General Public License v2.0"
__version__ = "2.0.0"
__maintainer__ = "Samuel Murail"
__email__ = "samuel.murail@u-paris.fr"
__status__ = "Prototype"

# Logging
logger = logging.getLogger(__name__)


def show_log(pdb_manip_log=True):

    from . import gmxsys
    gmxsys.show_log()



#!/usr/bin/env python3
# coding: utf-8
##################################
# #######   GROMACS 5   ##########
##################################


import os
import logging
import sys

from os_command_py import os_command

from .rtp import Rtp

# Logging
logger = logging.getLogger(__name__)

# Autorship information
__author__ = "Samuel Murail"
__copyright__ = "Copyright 2020, RPBS"
__credits__ = ["Samuel Murail"]
__license__ = "GNU General Public License v2.0"
__maintainer__ = "Samuel Murail"
__email__ = "samuel.murail@u-paris.fr"
__status__ = "Production"


def show_log():
    """ To use only with Doctest !!!
    Redirect logger output to sys.stdout
    """
    # Delete all handlers
    logger.handlers = []
    # Set the logger level to INFO
    logger.setLevel(logging.INFO)
    # Add sys.sdout as handler
    logger.addHandler(logging.StreamHandler(sys.stdout))


def show_debug():
    """ To use only with Doctest !!!
    Redirect logger output to sys.stdout
    """
    # Delete all handlers
    logger.handlers = []
    # Set the logger level to INFO
    logger.setLevel(logging.DEBUG)
    # Add sys.sdout as handler
    logger.addHandler(logging.StreamHandler(sys.stdout))


class TopMol:
    """Individual molecule topologie
    """

    def __init__(self, name, nrexcl):
        """Read an itp file and extract [atoms] field
        """
        self.name = name
        self.nrexcl = nrexcl
        self.atom_dict = {}
        self.bond_list = []
        self.cons_list = []
        self.pair_list = []
        self.angl_list = []
        self.dihe_list = []
        self.cmap_list = []
        self.vs3_list = []
        self.vs4_list = []
        self.pos_restr = []
        self.if_pos_restr = []

    def get_charge(self):
        self.charge = 0
        for atom in self.atom_dict.values():
            self.charge += atom['charge']
        return round(self.charge, 4)

    def get_res_num(self):
        self.res_num = -1
        for atom in self.atom_dict.values():
            if atom["res_num"] != self.res_num:
                self.res_num += 1
        return self.res_num

    def get_selection_index(self, selec_dict={'atom_name': ['CA']}):
        """Return the atom index to add posre
        """

        index_list = []
        for index, atom in self.atom_dict.items():
            selected = True
            for field, selection in selec_dict.items():
                if atom[field] not in selection:
                    selected = False
                    break
            if selected:
                index_list.append(index)

        return index_list

    def write_file(self, filout):
        filout.write("\n[ moleculetype ]\n; Name            nrexcl\n")
        filout.write("{}\t\t{}\n".format(self.name, self.nrexcl))
        # Print Atoms field
        filout.write("\n[ atoms ]\n;   nr       type  resnr residue  atom   "
                     "cgnr     charge       mass  typeB    chargeB      "
                     "massB\n")
        tot_charge = 0
        for atom in self.atom_dict.values():
            tot_charge += atom['charge']
            # print('ATOM:',atom)
            filout.write("{:>6}{:>11}{:>7}{:>7}{:>7}{:>7}{:>11.5f}{:>11} "
                         "  ; qtot {:<6.2f} \n".
                         format(atom['num'],
                                atom['atom_type'],
                                atom['res_num'],
                                atom['res_name'],
                                atom['atom_name'], atom['charge_num'],
                                atom['charge'],
                                atom['mass'],
                                tot_charge))
        # Print bonds field
        if self.bond_list:
            filout.write("\n[ bonds ]\n;  ai    aj funct            c0        "
                         "    c1            c2            c3\n")
            for param in self.bond_list:
                filout.write("{:>6}{:>6}{:>6}{:>13}{:>13}\n".format(
                    param['ai'], param['aj'], param['funct'],
                    param['r'], param['k']))
        # Print constraints field
        if self.cons_list:
            filout.write("\n[ constraints ]\n;  ai    aj funct            c0  "
                         "          c1\n")
            for param in self.cons_list:
                filout.write("{:>6}{:>6}{:>6}\n".format(
                    param['ai'], param['aj'], param['funct']))
        # Print pairs field
        if self.pair_list:
            filout.write("\n[ pairs ]\n;  ai    aj funct            c0        "
                         "    c1            c2            c3\n")
            for param in self.pair_list:
                filout.write("{:>6}{:>6}{:>6}\n".format(
                    param['ai'], param['aj'], param['funct']))
        # Print angles field
        if self.angl_list:
            filout.write("\n[ angles ]\n;  ai    aj    ak funct            c0 "
                         "           c1            c2            c3\n")
            for param in self.angl_list:
                filout.write("{:>6}{:>6}{:>6}{:>6}{:>13}{:>13}\n".format(
                    param['ai'], param['aj'],
                    param['ak'], param['funct'],
                    param['theta'], param['cth']))
        # Print dihedrals field
        if self.dihe_list:
            filout.write("\n[ dihedrals ]\n;  ai    aj    ak    al funct      "
                         "      c0            c1            c2            c3  "
                         "          c4            c5\n")
            for param in self.dihe_list:
                filout.write("{:>6}{:>6}{:>6}{:>6}{:>6}{:>13}{:>13}"
                             "{:>6}\n".format(param['ai'], param['aj'],
                                              param['ak'], param['al'],
                                              param['funct'], param['phase'],
                                              param['kd'], param['pn']))
        # Print virtual_sites3 field
        if self.cmap_list:
            filout.write("\n[ cmap ]\n;  ai    aj    ak    al    am funct\n")
            for param in self.cmap_list:
                filout.write("{:>6}{:>6}{:>6}{:>6}{:>6}{:>6}\n".format(
                    param['ai'], param['aj'],
                    param['ak'], param['al'],
                    param['am'], param['funct']))
        # Print virtual_sites3 field
        if self.vs3_list:
            filout.write("\n[ virtual_sites3 ]\n;  ai    aj    ak    al funct"
                         "            c0            c1\n")
            for param in self.vs3_list:
                filout.write("{:>6}{:>6}{:>6}{:>6}{:>6}\n".format(
                    param['ai'], param['aj'],
                    param['ak'], param['al'],
                    param['funct']))
        # Print virtual_sites3 field
        if self.vs4_list:
            filout.write("\n[ virtual_sites4 ]\n;  ai    aj    ak    al    am"
                         " funct            c0            c1            c2\n")
            for param in self.vs4_list:
                filout.write("{:>6}{:>6}{:>6}{:>6}{:>6}{:>6}\n".format(
                    param['ai'], param['aj'],
                    param['ak'], param['al'],
                    param['am'], param['funct']))
        # Print position restraints
        if self.pos_restr:
            filout.write("\n[ position_restraints ]\n;  ai    funct         "
                         "kx          ky          kz\n")
            for param in self.pos_restr:
                filout.write("{:>6}{:>6}{:>6}{:>6}{:>6}\n".format(
                    param['ai'], param['funct'],
                    param['kx'], param['ky'],
                    param['kz']))
        if len(self.if_pos_restr) > 0:
            def_str = ''
            def_end_flag = False
            for param in self.if_pos_restr:

                if param['def'] != def_str:
                    def_str = param['def']
                    if def_end_flag:
                        filout.write("#endif\n\n")

                    def_end_flag = True
                    filout.write("\n#ifdef " + param['def'] + "\n")
                    filout.write("\n[ position_restraints ]\n;  ai    funct "
                                 "        kx          ky          kz\n")

                filout.write("{:>6}{:>6}{:>6}{:>6}{:>6}\n".format(
                    param['ai'], param['funct'],
                    param['kx'], param['ky'],
                    param['kz']))
            filout.write("#endif\n\n")

    def delete_atom(self, index_list):
        # Remove atom:
        for i in index_list:
            del self.atom_dict[i]

        # Create the dict to have all atom num consecutive staring from 1
        dict_atom_index = {}
        for i, atom in sorted(enumerate(self.atom_dict.items())):
            # print(i, atom)
            dict_atom_index[atom[0]] = i + 1

        # Modify all atom to have correct index
        new_atom_dict = {}
        for i, atom in self.atom_dict.items():
            local_atom = atom
            new_index = dict_atom_index[local_atom['num']]
            local_atom['num'] = new_index
            local_atom['charge_num'] = new_index
            new_atom_dict[new_index] = local_atom

        self.atom_dict = new_atom_dict

        #  Modify all bond, cons, pair ... to have correct index:
        new_bond_list = []
        for i, param in enumerate(self.bond_list):
            if not ((param['ai'] in index_list) or
                    (param['aj'] in index_list)):
                param.update({'ai': dict_atom_index[param['ai']],
                              'aj': dict_atom_index[param['aj']]})
                new_bond_list.append(param)
        self.bond_list = new_bond_list

        new_cons_list = []
        for i, param in enumerate(self.cons_list):
            if not ((param['ai'] in index_list) or
                    (param['aj'] in index_list)):
                param.update({'ai': dict_atom_index[param['ai']],
                              'aj': dict_atom_index[param['aj']]})
                new_cons_list.append(param)
        self.cons_list = new_cons_list

        new_pair_list = []
        for i, param in enumerate(self.pair_list):
            if not ((param['ai'] in index_list) or
                    (param['aj'] in index_list)):
                param.update({'ai': dict_atom_index[param['ai']],
                              'aj': dict_atom_index[param['aj']]})
                new_pair_list.append(param)
        self.pair_list = new_pair_list

        new_angl_list = []
        for i, param in enumerate(self.angl_list):
            if not ((param['ai'] in index_list) or
                    (param['aj'] in index_list) or
                    (param['ak'] in index_list)):
                param.update({'ai': dict_atom_index[param['ai']],
                              'aj': dict_atom_index[param['aj']],
                              'ak': dict_atom_index[param['ak']]})
                new_angl_list.append(param)

        self.angl_list = new_angl_list

        new_dihe_list = []
        for i, param in enumerate(self.dihe_list):
            if not ((param['ai'] in index_list) or
                    (param['aj'] in index_list) or
                    (param['ak'] in index_list) or
                    (param['al'] in index_list)):
                param.update({'ai': dict_atom_index[param['ai']],
                              'aj': dict_atom_index[param['aj']],
                              'ak': dict_atom_index[param['ak']],
                              'al': dict_atom_index[param['al']]})
                new_dihe_list.append(param)

        self.dihe_list = new_dihe_list

        new_vs3_list = []
        for i, param in enumerate(self.vs3_list):
            if not ((param['ai'] in index_list) or
                    (param['aj'] in index_list) or
                    (param['ak'] in index_list) or
                    (param['al'] in index_list)):
                param.update({'ai': dict_atom_index[param['ai']],
                              'aj': dict_atom_index[param['aj']],
                              'ak': dict_atom_index[param['ak']],
                              'al': dict_atom_index[param['al']]})
                new_vs3_list.append(param)
        self.vs3_list = new_vs3_list

        new_cmap_list = []
        for i, param in enumerate(self.cmap_list):
            if not ((param['ai'] in index_list) or
                    (param['aj'] in index_list) or
                    (param['ak'] in index_list) or
                    (param['al'] in index_list) or
                    (param['am'] in index_list)):
                param.update({'ai': dict_atom_index[param['ai']],
                              'aj': dict_atom_index[param['aj']],
                              'ak': dict_atom_index[param['ak']],
                              'al': dict_atom_index[param['al']],
                              'am': dict_atom_index[param['am']]})
                new_cmap_list.append(param)
        self.cmap_list = new_cmap_list

        new_vs4_list = []
        for i, param in enumerate(self.vs4_list):
            if not ((param['ai'] in index_list) or
                    (param['aj'] in index_list) or
                    (param['ak'] in index_list) or
                    (param['al'] in index_list) or
                    (param['am'] in index_list)):
                param.update({'ai': dict_atom_index[param['ai']],
                              'aj': dict_atom_index[param['aj']],
                              'ak': dict_atom_index[param['ak']],
                              'al': dict_atom_index[param['al']],
                              'am': dict_atom_index[param['am']]})
                new_vs4_list.append(param)
        self.vs4_list = new_vs4_list

    def correct_charge_type(self, forcefield, index_list=None):
        """ Correct the charge and atom type of an itp object,
        base on a ff `.rtp` file.
        This is specially usefull, to correct charge of termini resiudes
        of a cyclicized peptide.

        if index_list is None, will correct all atom charges, if not, only
        atoms listed in index_list.
        """

        # First extract charge and type from the ff .rtp file
        forcefield_path = forcefield['path']
        rtp_path = os.path.abspath(os.path.join(
            os_command.get_directory(forcefield_path), 'aminoacids.rtp'))
        if not os_command.check_file_exist(rtp_path):
            rtp_path = os.path.abspath(os.path.join(
                os_command.get_directory(forcefield_path), 'merged.rtp'))

        logger.info('Read rtp file : {}'.format(rtp_path))
        ff_rtp = Rtp(rtp_path)

        # Correct charges and type:
        # for atom_num, atom in self.atom_dict.items():
        if index_list is None:
            index_list = self.atom_dict.keys()

        for atom_num in index_list:

            atom = self.atom_dict[atom_num]

            res_name = atom['res_name']
            resid = atom['res_num']
            atom_name = atom['atom_name']

            # With Amber disulfide cys are called CYX in the rtp and CYS in
            # the itp.
            # Don't get sense but need to check the protonation state of cys
            # By counting atoms
            # With charmm disuldie cys is CYS2
            if res_name == 'CYS':
                if len(self.get_selection_index(
                        selec_dict={'res_num': [resid],
                                    'res_name': ['CYS']})) == 10:
                    if forcefield['name'].startswith('amber'):
                        res_name = 'CYX'
                    elif forcefield['name'].startswith('charmm'):
                        res_name = 'CYS2'
            # print(res_name)
            # With gromacs all histidine are called HIS !
            if res_name == 'HIS':
                if len(self.get_selection_index(
                        selec_dict={'res_num': [resid],
                                    'res_name': ['HIS']})) == 18:
                    if forcefield['name'].startswith('amber'):
                        res_name = 'HIP'
                    elif forcefield['name'].startswith('charmm'):
                        res_name = 'HSP'
                elif len(self.get_selection_index(
                        selec_dict={'res_num': [resid],
                                    'atom_name': ['HE2']})) == 1:
                    if forcefield['name'].startswith('amber'):
                        res_name = 'HIE'
                    elif forcefield['name'].startswith('charmm'):
                        res_name = 'HSE'
                else:
                    if forcefield['name'].startswith('amber'):
                        res_name = 'HID'
                    elif forcefield['name'].startswith('charmm'):
                        res_name = 'HSD'

            # print(atom)
            # print(ff_rtp.res_dict[res_name]['atom'][atom_name])

            if atom_name in ff_rtp.res_dict[res_name]['atom']:
                atom_type = \
                    ff_rtp.res_dict[res_name]['atom'][atom_name]['type']
                atom_charge = \
                    ff_rtp.res_dict[res_name]['atom'][atom_name]['charge']
                # print('Correct residue {:4} atom {:4} atom type {:4} '
                #       'to {:4}'.format(res_name, atom['atom_name'],
                #                        atom['atom_type'], atom_type))
                # print('Correct charge {:4} '
                #       'to {:4}'.format(self.atom_dict[atom_num]['charge'],
                #                        atom_charge))
                if atom_type != atom['atom_type']:
                    logger.warning('Correct residue {:4} atom {:4} atom '
                                   'type {:4} to {:4}'.format(
                                        res_name, atom['atom_name'],
                                        atom['atom_type'], atom_type))
            else:
                logger.warning('Can\'t find residue {:4} atom {:4} atom '
                               'type {:4} parameters in forcefield'.format(
                                    res_name, atom['atom_name'],
                                    atom['atom_type']))
            self.atom_dict[atom_num]['atom_type'] = atom_type
            self.atom_dict[atom_num]['charge'] = atom_charge



#!/usr/bin/env python3

# coding: utf-8

""" Collection of function to use the antechamber toolbox.

https://ambermd.org/AmberTools.php

"""

import os
import logging
import sys

from os_command_py import os_command
from pdb_manip_py import pdb_manip

from .. import gmx

# Logging
logger = logging.getLogger(__name__)


# Autorship information
__author__ = "Samuel Murail"
__copyright__ = "Copyright 2020, RPBS"
__credits__ = ["Samuel Murail"]
__license__ = "GNU General Public License v2.0"
__maintainer__ = "Samuel Murail"
__email__ = "samuel.murail@u-paris.fr"
__status__ = "Production"

# Test folder path
MONITOR_LIB_DIR = os.path.dirname(os.path.abspath(__file__))
TEST_PATH = os.path.abspath(os.path.join(MONITOR_LIB_DIR, "../test_files/"))

# Check if Readthedoc is launched skip the program path searching
on_rtd = os.environ.get('READTHEDOCS') == 'True'
if on_rtd:
    logger.info("Ambertools cannot be found")
    REDUCE_BIN = ""
    ANTECHAMBER_BIN = ""
else:
    # As the mabertools module is not a hard dependency
    # gromacs_py should be loaded even if ambertools 
    # is note installes
    try:
        REDUCE_BIN = os_command.which('reduce')
        ANTECHAMBER_BIN = os_command.which('antechamber')
        ACPYPE_BIN = os_command.which('acpype', 'acpype.py')
    except IOError:
        logger.warning('Reduce or antechamber or acpype binary cannot be'
                       ' found. Install it using conda \n conda install '
                       'acpype')


def show_log():
    """ To use only with Doctest !!!
    Redirect logger output to sys.stdout
    """
    # Delete all handlers
    logger.handlers = []
    # Set the logger level to INFO
    logger.setLevel(logging.INFO)
    # Add sys.sdout as handler
    logger.addHandler(logging.StreamHandler(sys.stdout))


def show_debug(pdb_manip_log=True):
    """ To use only with Doctest !!!
    Redirect logger output to sys.stdout
    """
    # Delete all handlers
    logger.handlers = []
    # Set the logger level to INFO
    logger.setLevel(logging.DEBUG)
    # Add sys.sdout as handler
    logger.addHandler(logging.StreamHandler(sys.stdout))


def smile_to_pdb(smile, pdb_out, mol_name,
                 method_3d='rdkit', iter_num=5000):
    """
    """

    if method_3d == 'openbabel':

        from openbabel import pybel

        conf = pybel.readstring("smi", smile)
        # Get charge
        charge = conf.charge
        conf.make3D(forcefield='mmff94', steps=iter_num)
        conf.write(format='pdb', filename=pdb_out, overwrite=True)

    elif method_3d == 'rdkit':

        from rdkit.Chem import AllChem as Chem

        conf = Chem.MolFromSmiles(smile)
        conf = Chem.AddHs(conf)
        # Get charge
        charge = Chem.GetFormalCharge(conf)
        Chem.EmbedMolecule(conf)
        Chem.MMFFOptimizeMolecule(
            conf, mmffVariant='MMFF94', maxIters=iter_num)
        Chem.MolToPDBFile(conf, filename=pdb_out)

    # Change resname of pdb file to `self.mol_name`
    coor = pdb_manip.Coor(pdb_out)
    index_list = coor.get_index_selection(selec_dict={'res_name': ['UNL']})
    coor.change_index_pdb_field(index_list, change_dict={
        'res_name': mol_name})
    coor.write_pdb(pdb_out, check_file_out=False)

    return(charge)


def add_hydrogen(pdb_in, pdb_out, check_file_out=True, **reduce_options):
    """Add hydrogen to a pdb file using the ``reduce`` software:

    :param pdb_in: pdb input
    :type pdb_in: str

    :param pdb_out: pdb output
    :type pdb_out: str

    :param check_file_out: flag to check or not if file has already been
        created. If the file is present then the command break.
    :type check_file_out: bool, optional, default=True

    :param reduce_options: Optional arguments for ``reduce``


    :Example:

    >>> from pdb_manip_py import pdb_manip
    >>> pdb_manip.show_log()
    >>> TEST_OUT = getfixture('tmpdir')
    >>> add_hydrogen(pdb_in=TEST_PATH+'/phenol.pdb',\
    pdb_out=TEST_OUT+'/phenol_h.pdb') #doctest: +ELLIPSIS
    reduce -build -nuclear .../phenol.pdb
    >>> phenol_coor = pdb_manip.Coor(TEST_OUT+'/phenol_h.pdb')  \
#doctest: +ELLIPSIS
    Succeed to read file .../phenol_h.pdb ,  13 atoms found

    """
    # Check if output files exist:
    if check_file_out and os.path.isfile(pdb_out):
        logger.info("PDB files not created, {} already exist".format(
            pdb_out))
        return

    # Define reduce command:
    cmd_reduce = os_command.Command([REDUCE_BIN,
                                     "-build",
                                     "-nuclear",
                                     pdb_in], **reduce_options)

    cmd_reduce.display()
    return_code = cmd_reduce.run(display=False, out_data=True)

    filout = open(pdb_out, 'w')
    filout.write(return_code['stdout'])

    logger.info("Succeed to save file %s" % os.path.relpath(pdb_out))

    return


def add_hydrogen_rdkit(pdb_in, smile, pdb_out):
    """Add hydrogen to a pdb file using the ``rdkit`` library:

    :param pdb_in: pdb input
    :type pdb_in: str

    :param pdb_out: pdb output
    :type pdb_out: str

    :Example:

    >>> from pdb_manip_py import pdb_manip
    >>> pdb_manip.show_log()
    >>> TEST_OUT = getfixture('tmpdir')
    >>> # print(TEST_OUT)
    >>> add_hydrogen_rdkit(pdb_in=os.path.join(TEST_PATH,'four_phenol.pdb'),\
smile="C1=CC=C(C=C1)O",\
pdb_out=os.path.join(TEST_OUT,'four_phenol_h.pdb')) #doctest: +ELLIPSIS
    Succeed to read file ...four_phenol.pdb ,  28 atoms found
    Succeed to save file ...four_phenol_0.pdb
    Succeed to read file ...four_phenol_0.pdb ,  7 atoms found
    Succeed to read file ...four_phenol_0_h.pdb ,  13 atoms found
    Succeed to save file ...four_phenol_0_h.pdb
    Succeed to read file ...four_phenol_0_h.pdb ,  13 atoms found
    Succeed to save file ...four_phenol_0_h.pdb
    Succeed to save file ...four_phenol_1.pdb
    Succeed to read file ...four_phenol_1.pdb ,  7 atoms found
    Succeed to read file ...four_phenol_1_h.pdb ,  13 atoms found
    Succeed to save file ...four_phenol_1_h.pdb
    Succeed to read file ...four_phenol_1_h.pdb ,  13 atoms found
    Succeed to save file ...four_phenol_1_h.pdb
    Succeed to save file ...four_phenol_2.pdb
    Succeed to read file ...four_phenol_2.pdb ,  7 atoms found
    Succeed to read file ...four_phenol_2_h.pdb ,  13 atoms found
    Succeed to save file ...four_phenol_2_h.pdb
    Succeed to read file ...four_phenol_2_h.pdb ,  13 atoms found
    Succeed to save file ...four_phenol_2_h.pdb
    Succeed to save file ...four_phenol_3.pdb
    Succeed to read file ...four_phenol_3.pdb ,  7 atoms found
    Succeed to read file ...four_phenol_3_h.pdb ,  13 atoms found
    Succeed to save file ...four_phenol_3_h.pdb
    Succeed to read file ...four_phenol_3_h.pdb ,  13 atoms found
    Succeed to save file ...four_phenol_3_h.pdb
    Succeed to save concat file:  ...four_phenol_h.pdb
    0
    >>> phenol_coor = pdb_manip.Coor(os.path.join(TEST_OUT,'four_phenol_h.pdb'))\
#doctest: +ELLIPSIS
    Succeed to read file ...four_phenol_h.pdb ,  52 atoms found

    """

    full_coor = pdb_manip.Coor(pdb_in)

    # Change first residue to 1, as it is the res from rdkit output
    res_list = full_coor.get_attribute_selection(attribute='res_num')
    pdb_list = []
    pdb_del_list = []

    for i, res in enumerate(res_list):
        # Extract residue
        one_res = full_coor.select_part_dict(
            selec_dict={'res_num': [res]})
        out_pdb = '{}_{}.pdb'.format(pdb_in[:-4], i)
        one_res.write_pdb(out_pdb)
        pdb_del_list.append(out_pdb)

        # Add hydrogens
        out_h_pdb = '{}_{}_h.pdb'.format(pdb_in[:-4], i)
        charge = add_hydrogen_rdkit_one_mol(out_pdb, smile, out_h_pdb)
        pdb_list.append(out_h_pdb)
        pdb_del_list.append(out_h_pdb)

        # Update residue number:
        one_res_h = pdb_manip.Coor(out_h_pdb)
        one_res_h.change_pdb_field(
            change_dict={"res_num": i + 1})
        one_res_h.write_pdb(out_h_pdb, check_file_out=False)


    pdb_manip.Coor.concat_pdb(*pdb_list,
                              pdb_out = pdb_out)

    # Delete intermediate files
    for pdb in pdb_del_list:
        os_command.delete_file(pdb)

    return charge

def add_hydrogen_rdkit_one_mol(pdb_in, smile, pdb_out):
    """Add hydrogen to a pdb file using the ``rdkit`` library:

    :param pdb_in: pdb input
    :type pdb_in: str

    :param pdb_out: pdb output
    :type pdb_out: str

    :Example:

    >>> from pdb_manip_py import pdb_manip
    >>> pdb_manip.show_log()
    >>> TEST_OUT = getfixture('tmpdir')
    >>> # print(TEST_OUT)
    >>> add_hydrogen_rdkit_one_mol(pdb_in=os.path.join(TEST_PATH,'phenol.pdb'),\
smile="C1=CC=C(C=C1)O",\
pdb_out=os.path.join(TEST_OUT,'phenol_h.pdb')) #doctest: +ELLIPSIS
    Succeed to read file ...phenol_h.pdb ,  13 atoms found
    Succeed to save file ...phenol_h.pdb
    0
    >>> phenol_coor = pdb_manip.Coor(os.path.join(TEST_OUT,'phenol_h.pdb'))\
#doctest: +ELLIPSIS
    Succeed to read file .../phenol_h.pdb ,  13 atoms found

    .. warning:

        

    """

    try:
        from rdkit.Chem import AllChem as Chem
        # from rdkit.Chem import rdMolAlign as align
    except ImportError:
        logger.error('Could not load rdkit \nInstall it using conda:\n'
                     'conda install -c conda-forge rdkit')
        sys.exit(1)

    lig_pdb = Chem.MolFromPDBFile(pdb_in, removeHs=True)
    lig_smile = Chem.MolFromSmiles(smile)

    # Need to count the number of molecule
    pdb_atom_num = lig_pdb.GetNumAtoms()
    smile_atom_num = lig_smile.GetNumAtoms()

    # If more than one molecule, add them
    # in the smile string
    if pdb_atom_num != smile_atom_num:
        mol_num = pdb_atom_num/smile_atom_num
        smile_list = [smile] * int(mol_num)
        lig_smile = Chem.MolFromSmiles('.'.join(smile_list))

    # Assign bond order on pdb using smile informations
    newMol = Chem.AssignBondOrdersFromTemplate(lig_smile, lig_pdb)

    # Add hydrogens to lig_smile, using lig_pdb as coordinates constraints
    # This way is better than adding h to lig_pdb
    # because acpype experience some issue
    lig_smile_h = Chem.AddHs(lig_smile)
    Chem.ConstrainedEmbed(lig_smile_h, newMol)
    Chem.MolToPDBFile(lig_smile_h, pdb_out)

    # OLD WAY :
    # Add hydrogens
    # newMol_h = Chem.AddHs(newMol)

    # # Need to define how to match atoms form pdb to smile
    # match_atom = newMol_h.GetSubstructMatch(lig_smile)
    # cmap = {match_atom[i]: lig_pdb.GetConformer().
    #         GetAtomPosition(match_atom[i]) for i in range(len(match_atom))}

    # # Hydrogens coordinates need to be computed
    # # While keeping heavy atoms coordinates
    # Chem.EmbedMolecule(newMol_h, coordMap=cmap)

    # # Align new coordinates to old one
    # align.AlignMol(newMol_h, newMol,
    #                atomMap=[[i, i] for i in range(len(match_atom))])
    #  Save coordinates
    # Chem.MolToPDBFile(newMol_h, pdb_out)

    # Change UNL residue name to original one
    coor_start = pdb_manip.Coor(pdb_in)
    res_name_list_start = coor_start.get_attribute_selection(
        attribute='res_name')
    if len(res_name_list_start) > 1:
        res_name_list_start.remove('UNL')

    coor = pdb_manip.Coor(pdb_out)
    index_list = coor.get_index_selection(selec_dict={'res_name': ['UNL']})
    coor.change_index_pdb_field(index_list, change_dict={
        'res_name': res_name_list_start[0]})
    coor.write_pdb(pdb_out, check_file_out=False)

    # Return charge
    return Chem.GetFormalCharge(lig_smile)

    # Need to fix the residue number


def antechamber(pdb_in, mol2_out, charge_model="bcc",
                check_file_out=True, **antechamber_options):
    """Compute a molecule topologie using the ``antechamber`` software:

    :param pdb_in: pdb input
    :type pdb_in: str

    :param mol2_out: pdb output
    :type mol2_out: str

    :param charge_model: charge model
    :type charge_model: str, default="bcc"

    :param check_file_out: flag to check or not if file has already been
        created. If the file is present then the command break.
    :type check_file_out: bool, optional, default=True

    :param reduce_options: Optional arguments for ``reduce``

    Output files:

        - mol2_out
        - ANTECHAMBER_AM1BCC_PRE.AC
        - ANTECHAMBER_BOND_TYPE.AC
        - ANTECHAMBER_BOND_TYPE.AC0
        - ANTECHAMBER_AC.AC
        - ATOMTYPE.INF
        - ANTECHAMBER_AC.AC0
        - ANTECHAMBER_AM1BCC.AC

    :Example:

    >>> pdb_manip.show_log()
    >>> TEST_OUT = getfixture('tmpdir')
    >>> add_hydrogen(pdb_in=TEST_PATH+'/phenol.pdb',\
    pdb_out=TEST_OUT+'/phenol_h.pdb') #doctest: +ELLIPSIS
    reduce -build -nuclear .../phenol.pdb
    >>> phenol_coor = pdb_manip.Coor(TEST_OUT+'/phenol_h.pdb')  \
#doctest: +ELLIPSIS
    Succeed to read file .../phenol_h.pdb ,  13 atoms found
    >>> antechamber(pdb_in=TEST_OUT+'/phenol_h.pdb',\
    mol2_out=TEST_OUT+'/phenol_h.mol2') #doctest: +ELLIPSIS
    antechamber -i phenol_h.pdb -fi pdb -o phenol_h.mol2 -fo \
mol2 -c bcc

    """

    # Check if output files exist:
    if check_file_out and os.path.isfile(mol2_out):
        logger.info("MOL2 files not created, {} already exist".format(
            mol2_out))
        return

    start_dir = os.path.abspath(".")
    # Create and go in out_folder:
    out_folder = os_command.get_directory(mol2_out)
    os_command.create_and_go_dir(out_folder)

    # Define reduce command:
    cmd_antechamber = os_command.Command([ANTECHAMBER_BIN,
                                          "-i", pdb_in,
                                          "-fi", str(pdb_in).split('.')[-1],
                                          "-o", mol2_out,
                                          "-fo", "mol2",
                                          "-c", charge_model],
                                         **antechamber_options)

    cmd_antechamber.display()
    cmd_antechamber.run(display=False)

    logger.info("Succeed to save file %s" % os.path.relpath(mol2_out))

    # Get absolute path:
    os.chdir(start_dir)

    return


def acpype(pdb_in, out_folder, charge_model="bcc",
           atom_type="gaff", net_charge=None,
           check_file_out=True, **acpype_options):
    """Compute a molecule topologie using the ``antechamber`` software:

    :param pdb_in: pdb input
    :type pdb_in: str

    :param out_folder: output folder
    :type out_folder: str

    :param charge_model: charge model
    :type charge_model: str, default="bcc"

    :param atom_type: atom_type model
    :type atom_type: str, default="gaff"

    :param check_file_out: flag to check or not if file has already been
        created. If the file is present then the command break.
    :type check_file_out: bool, optional, default=True

    :param acpype_options: Optional arguments for ``acpype``


    :Example:

    >>> pdb_manip.show_log()
    >>> TEST_OUT = getfixture('tmpdir')
    >>> add_hydrogen(pdb_in=TEST_PATH+'/phenol.pdb',\
    pdb_out=TEST_OUT+'/phenol_h.pdb') #doctest: +ELLIPSIS
    reduce -build -nuclear .../phenol.pdb
    >>> phenol_coor = pdb_manip.Coor(TEST_OUT+'/phenol_h.pdb')  \
#doctest: +ELLIPSIS
    Succeed to read file .../phenol_h.pdb ,  13 atoms found

    """

    # Check if output files exist:
    if check_file_out and os.path.isfile(out_folder):
        logger.info("MOL2 files not created, {} already exist".format(
            out_folder))
        return

    # Remove tha last char if == '/'
    if out_folder[-1] == '/':
        out_folder = out_folder[:-1]

    name = out_folder.split('/')[-1]

    cmd_list = [ACPYPE_BIN,
                "-i", pdb_in,
                "-b", out_folder,
                "-c", charge_model,
                "-a", atom_type,
                "-o", "gmx"]

    if net_charge is not None:
        cmd_list += ["-n", str(net_charge)]

    # Define reduce command:
    cmd_acpype = os_command.Command(cmd_list, **acpype_options)

    cmd_acpype.display()
    cmd_acpype.run(display=False)

    logger.info("Succeed to create topologie in %s" % os.path.relpath(
        out_folder))

    # Split the itp file atom_type part:
    extract_itp_atomtypes('{}.acpype/{}_GMX.itp'.format(
                          out_folder, name),
                          '{}.acpype/{}_GMX_atomtypes.itp'.format(
                          out_folder, name))

    sys_top = gmx.TopSys('{}.acpype/{}_GMX.top'.format(out_folder, name))

    fullname = '{}_GMX_atomtypes.itp'.format(name)
    include = '{}_GMX_atomtypes'.format(name)
    path = '{}.acpype/{}_GMX_atomtypes.itp'.format(out_folder, name)

    atomtypes_itp = gmx.Itp(name=include, fullname=fullname, path=path)

    sys_top.itp_list = [atomtypes_itp] + sys_top.itp_list
    # Add position restraints
    # Get heavy atoms name:
    atom_list = []
    for key, value in sys_top.itp_list[-1].top_mol_list[0].atom_dict.items():
        if not value['atom_name'].startswith('H'):
            atom_list.append(value['atom_name'])

    sys_top.add_posre(posre_name="POSRES", selec_dict={
            'atom_name': atom_list},
            fc=[1000, 1000, 1000])
    sys_top.add_posre(posre_name="POSRES_LIG", selec_dict={
            'atom_name': atom_list},
            fc=[1000, 1000, 1000])

    sys_top.add_posre(posre_name="POSRES_CA", selec_dict={
            'atom_name': atom_list},
            fc=[1000, 1000, 1000])

    sys_top.add_posre(posre_name="POSRES_CA_LOW", selec_dict={
            'atom_name': atom_list},
            fc=[100, 100, 100])

    sys_top.write_file('{}.acpype/{}_GMX_split.top'.format(
                                out_folder, name))

    molecule = gmx.GmxSys(name='mol',
                          coor_file='{}.acpype/{}_GMX.gro'.format(
                            out_folder, name),
                          top_file='{}.acpype/{}_GMX_split.top'.format(
                            out_folder, name))

    return molecule


def make_amber_top_mol(pdb_in, res_name, charge, charge_model="bcc",
                       atom_type="gaff", remove_h=True):

    full_coor = pdb_manip.Coor(pdb_in)
    # Select coor:
    mol_coor = full_coor.select_part_dict(selec_dict={'res_name': [res_name]})

    # Remove hydrogens:
    if remove_h:
        to_del_list = []
        for atom_num, atom in mol_coor.atom_dict.items():
            if atom['name'][0] == 'H':
                to_del_list.append(atom_num)
        mol_coor.del_atom_index(to_del_list)

    mol_coor.write_pdb(res_name + '.pdb')

    # Add hydrogens:
    add_hydrogen(res_name + '.pdb', res_name + '_h.pdb')

    # Get only one molecule
    mol_h_coor = pdb_manip.Coor(res_name + '_h.pdb')
    res_list = mol_h_coor.get_attribute_selection(attribute='uniq_resid')
    mol_uniq_coor = mol_h_coor.select_part_dict(
        selec_dict={'uniq_resid': [res_list[0]]})
    # Save coordinates:
    mol_uniq_coor.write_pdb(res_name + '_h_unique.pdb')

    # Compute topologie with acpype:
    gmxsys = acpype(res_name + '_h_unique.pdb', res_name,
                    net_charge=charge,
                    charge_model="bcc",
                    atom_type="gaff")

    return({'GmxSys': gmxsys,
            'coor': os_command.full_path_and_check(res_name + '_h.pdb'),
            'num': len(res_list)})


def make_amber_top_mol_rdkit(pdb_in, res_name, smile, charge_model="bcc",
                             atom_type="gaff", remove_h=True):

    full_coor = pdb_manip.Coor(pdb_in)
    # Select coor:
    mol_coor = full_coor.select_part_dict(selec_dict={'res_name': [res_name]})

    # Change first residue to 1, as it is the res from rdkit output
    res_list = mol_coor.get_attribute_selection(attribute='res_num')
    index_all_list = []
    for res in res_list:
        index_list = mol_coor.get_index_selection(
            selec_dict={'res_num': [res]})
        index_all_list.append(index_list)

    for i, index_list in enumerate(index_all_list):
        mol_coor.change_index_pdb_field(index_list, change_dict={'res_num': i+1})

    #index_list = mol_coor.get_index_selection(
    #    selec_dict={'res_num': [res_list[0]]})
    #mol_coor.change_index_pdb_field(index_list, change_dict={'res_num': 1})

    # Remove hydrogens:
    if remove_h:
        to_del_list = []
        for atom_num, atom in mol_coor.atom_dict.items():
            if atom['name'].startswith('H'):
                to_del_list.append(atom_num)
        mol_coor.del_atom_index(to_del_list)

    mol_coor.write_pdb(res_name + '.pdb', check_file_out=False)

    # Add hydrogens:
    # add_hydrogen(res_name+'.pdb', res_name+'_h.pdb')
    charge = add_hydrogen_rdkit(res_name + '.pdb',
                                smile,
                                res_name + '_h.pdb')

    # Get only one molecule
    mol_h_coor = pdb_manip.Coor(res_name + '_h.pdb')
    res_list = mol_h_coor.get_attribute_selection(attribute='uniq_resid')
    mol_uniq_coor = mol_h_coor.select_part_dict(
        selec_dict={'uniq_resid': [res_list[0]]})
    # Save coordinates:
    mol_uniq_coor.write_pdb(res_name + '_h_unique.pdb', check_file_out=False)

    # Compute topologie with acpype:
    gmxsys = acpype(res_name+'_h_unique.pdb', res_name,
                    net_charge=charge,
                    charge_model="bcc",
                    atom_type="gaff")

    return({'GmxSys': gmxsys,
            'coor': os_command.full_path_and_check(res_name + '_h.pdb'),
            'num': len(res_list)})


def extract_itp_atomtypes(itp_in, itp_atomtypes_out):

    field = None
    atom_types_list = "[ atomtypes ]\n"

    with open(itp_in) as itpfile:
        for line in itpfile:

            if line.startswith("["):
                # Remove space and [ ]
                field = line.strip()[1:-1].strip()
                # print(field)
                continue
            if field == 'atomtypes':
                atom_types_list += line

    filout = open(itp_atomtypes_out, "w")
    filout.write(atom_types_list)
    filout.close()

    # Read and write itp to remove the [ atomtypes ] part
    fullname = (itp_in.split("/")[-1])
    include = fullname.split(".")[0]
    path = os_command.full_path_and_check(itp_in)

    top_mol = gmx.Itp(name=include, fullname=fullname, path=path)
    top_mol.write_file(path)



#!/usr/bin/env python3
# coding: utf-8

"""
Tests for GmxSys class
"""

import pytest
import os

from gromacs_py import gmx
import gromacs_py.tools.ambertools as ambertools

from pdb_manip_py import pdb_manip

from .datafiles import PDB_1Y0M, PDB_1RXZ, PDB_1JD4

# Autorship information
__author__ = "Samuel Murail"
__copyright__ = "Copyright 2020, RPBS"
__credits__ = ["Samuel Murail"]
__license__ = "GNU General Public License v2.0"
__maintainer__ = "Samuel Murail"
__email__ = "samuel.murail@u-paris.fr"
__status__ = "Production"


def test_zinc_finger(tmp_path):

    gmx.show_log()
    ##################################
    # ##   Create the topologie:   ###
    ##################################
    prot = gmx.GmxSys(name='1jd4', coor_file=PDB_1JD4)
    prot.prepare_top(out_folder=os.path.join(tmp_path, 'top_DIAP1'),
                     vsite='hydrogens', ignore_ZN=False)

    top_coor = pdb_manip.Coor(prot.coor_file)
    assert top_coor.num == 3292

    prot.switch_ion_octa_dummy()
    top_coor = pdb_manip.Coor(prot.coor_file)
    assert top_coor.num == 3304

    prot.solvate_add_ions(out_folder=os.path.join(tmp_path, 'top_sys'),
                          maxwarn=3)

    prot.em(out_folder=os.path.join(tmp_path, 'em_DIAP1'),
            nsteps=10,
            constraints='none')
    em_coor = pdb_manip.Coor(prot.coor_file)
    assert em_coor.num == 56366

    ener_pd = prot.get_ener(['Potential'])
    assert ener_pd['Potential'].values[-1] < -155000


def test_set_protonate_charmm(tmp_path):

    gmx.show_log()
    ##########################################
    # ##   Create the topologie at PH 4.0  ###
    ##########################################

    res_prot_dict = {'GLUP': {'res_num': [9, 24, 31, 57]},
                     'ASPP': {'res_num': [3, 70]},
                     'HSP': {'res_num': [28, 35]},
                     'HSD': {'res_num': [175]},
                     'HSE': {'res_num': [214]}}

    prot = gmx.GmxSys(name='1RXZ_ph4', coor_file=PDB_1RXZ)
    prot.prepare_top(out_folder=os.path.join(tmp_path, 'top_1RXZ_ph4'), ph=7.0,
                     res_prot_dict=res_prot_dict)
    top_coor = pdb_manip.Coor(prot.coor_file)
    assert top_coor.num == 4081

    top_1RXZ = gmx.TopSys(prot.top_file)
    assert top_1RXZ.charge() == -4


def test_set_protonate_amber(tmp_path):

    gmx.show_log()
    ##########################################
    # ##   Create the topologie at PH 4.0  ###
    ##########################################

    res_prot_dict = {'GLH': {'res_num': [9, 24, 31, 57]},
                     'ASH': {'res_num': [3, 70]},
                     'HIP': {'res_num': [28, 35]},
                     'HID': {'res_num': [175]},
                     'HIE': {'res_num': [214]}}

    prot = gmx.GmxSys(name='1RXZ_ph4', coor_file=PDB_1RXZ)
    prot.prepare_top(out_folder=os.path.join(tmp_path, 'top_1RXZ_ph4'), ph=7.0,
                     res_prot_dict=res_prot_dict,
                     ff="amber99sb-ildn")
    top_coor = pdb_manip.Coor(prot.coor_file)
    assert top_coor.num == 4081

    top_1RXZ = gmx.TopSys(prot.top_file)
    assert top_1RXZ.charge() == -4


def test_protonate_ph4_7_12_charmm(tmp_path):

    gmx.show_log()
    ##########################################
    # ##   Create the topologie at PH 4.0  ###
    ##########################################
    prot = gmx.GmxSys(name='1RXZ_ph4', coor_file=PDB_1RXZ)
    prot.prepare_top(out_folder=os.path.join(tmp_path, 'top_1RXZ_ph4'), ph=4.0)
    top_coor = pdb_manip.Coor(prot.coor_file)
    assert top_coor.num == 4108

    top_1RXZ = gmx.TopSys(prot.top_file)
    assert top_1RXZ.charge() == 23

    ##########################################
    # ##   Create the topologie at PH 7.0  ###
    ##########################################
    prot = gmx.GmxSys(name='1RXZ_ph7', coor_file=PDB_1RXZ)
    prot.prepare_top(out_folder=os.path.join(tmp_path, 'top_1RXZ_ph7'), ph=7.0)
    top_coor = pdb_manip.Coor(prot.coor_file)
    assert top_coor.num == 4073

    top_1RXZ = gmx.TopSys(prot.top_file)
    assert top_1RXZ.charge() == -12

    ##########################################
    # ##   Create the topologie at PH 10.0 ###
    ##########################################

    # Remark: Charmm doesn't allow lysine unprotonated
    prot = gmx.GmxSys(name='1RXZ_ph10', coor_file=PDB_1RXZ)
    prot.prepare_top(out_folder=os.path.join(tmp_path, 'top_1RXZ_ph10'),
                     ph=10.0)
    top_coor = pdb_manip.Coor(prot.coor_file)
    assert top_coor.num == 4073

    top_1RXZ = gmx.TopSys(prot.top_file)
    assert top_1RXZ.charge() == -12


def test_protonate_ph4_7_amber(tmp_path):

    gmx.show_log()
    ##########################################
    # ##   Create the topologie at PH 4.0  ###
    ##########################################
    prot = gmx.GmxSys(name='1RXZ_ph4', coor_file=PDB_1RXZ)
    prot.prepare_top(out_folder=os.path.join(tmp_path, 'top_1RXZ_ph4'),
                     ph=4.0, ff="amber99sb-ildn")
    top_coor = pdb_manip.Coor(prot.coor_file)
    assert top_coor.num == 4108

    top_1RXZ = gmx.TopSys(prot.top_file)
    assert top_1RXZ.charge() == 23

    ##########################################
    # ##   Create the topologie at PH 7.0  ###
    ##########################################
    prot = gmx.GmxSys(name='1RXZ_ph7', coor_file=PDB_1RXZ)
    prot.prepare_top(out_folder=os.path.join(tmp_path, 'top_1RXZ_ph7'),
                     ph=7.0, ff="amber99sb-ildn")
    top_coor = pdb_manip.Coor(prot.coor_file)
    assert top_coor.num == 4073

    top_1RXZ = gmx.TopSys(prot.top_file)
    assert top_1RXZ.charge() == -12

    ##########################################
    # ##   Create the topologie at PH 10.0 ###
    ##########################################
    # Remark: Amber allow lysine unprotonated
    prot = gmx.GmxSys(name='1RXZ_ph10', coor_file=PDB_1RXZ)
    prot.prepare_top(out_folder=os.path.join(tmp_path, 'top_1RXZ_ph10'),
                     ph=10.0, ff="amber99sb-ildn")
    top_coor = pdb_manip.Coor(prot.coor_file)
    assert top_coor.num == 4071

    top_1RXZ = gmx.TopSys(prot.top_file)
    assert top_1RXZ.charge() == -14


def test_insert_ethanol(tmp_path):

    gmx.show_log()
    ##################################
    # ##   Create the topologie:   ###
    ##################################
    prot = gmx.GmxSys(name='1y0m', coor_file=PDB_1Y0M)
    prot.prepare_top(out_folder=os.path.join(tmp_path, 'top_SH3'))
    top_coor = pdb_manip.Coor(prot.coor_file)
    assert top_coor.num == 996

    prot.solvate_add_ions(out_folder=os.path.join(tmp_path, 'top_sys'))

    prot.em(out_folder=os.path.join(tmp_path, 'em_SH3'),
            nsteps=10,
            constraints='none')
    em_coor = pdb_manip.Coor(prot.coor_file)
    assert em_coor.num == 15383

    ener_pd = prot.get_ener(['Potential'])
    assert ener_pd['Potential'].values[-1] < -155000

    ###################################
    # ###   Create Ethanol system   ###
    ###################################
    smile = 'CCO'
    eth_pdb = os.path.join(tmp_path, 'ethanol.pdb')
    charge = ambertools.smile_to_pdb(smile, eth_pdb, 'ETH')

    assert charge == 0

    eth_sys = gmx.GmxSys(name='ETH', coor_file=eth_pdb)
    eth_sys.prepare_top_ligand(
        out_folder=os.path.join(tmp_path, 'eth_top'),
        ff='amber99sb-ildn', include_mol={'ETH': smile})

    eth_sys.create_box(name=None, dist=0.5)

    ####################################################
    # # Insert 4 copy of ethanol in  the SH3 system: ###
    ####################################################

    prot.insert_mol_sys(mol_gromacs=eth_sys, mol_num=10,
                        new_name='SH3_ETH',
                        out_folder=os.path.join(tmp_path, 'top_ETH_SH3'))
    ################################
    # ##   Minimize the system   ###
    ################################
    prot.em_2_steps(out_folder=os.path.join(tmp_path, 'em_ETH_SH3'),
                    no_constr_nsteps=10, constr_nsteps=10)
    prot_lig_coor = pdb_manip.Coor(prot.coor_file)
    assert 15300 < prot_lig_coor.num < 15518

    prot.display_history()


@pytest.mark.skipif(float(gmx.gmxsys.gmx_version) >= 2019,
                    reason="Gromacs verions >= 19 have issues with ACE")
def test_insert_peptide_vsite(tmp_path):

    gmx.show_log()
    ##################################
    # ##   Create the topologie:   ###
    ##################################
    prot = gmx.GmxSys(name='1y0m', coor_file=PDB_1Y0M)
    prot.prepare_top(out_folder=os.path.join(tmp_path, 'top_SH3'),
                     vsite='hydrogens')
    top_coor = pdb_manip.Coor(prot.coor_file)
    assert top_coor.num == 1064

    prot.solvate_add_ions(out_folder=os.path.join(tmp_path, 'top_sys'))

    prot.em(out_folder=os.path.join(tmp_path, 'em_SH3'),
            nsteps=10,
            constraints='none')
    em_coor = pdb_manip.Coor(prot.coor_file)
    assert em_coor.num == 15337

    ener_pd = prot.get_ener(['Potential'])
    assert ener_pd['Potential'].values[-1] < -155000

    ###################################
    # ##    Create a DD peptide     ###
    ###################################
    pep = gmx.GmxSys(name='DD')
    pep.create_peptide(sequence='DD',
                       out_folder=os.path.join(tmp_path, 'top_DD'),
                       em_nsteps=10, equi_nsteps=0,
                       vsite='hydrogens')

    ####################################################
    # # Insert 4 copy of ethanol in  the SH3 system: ###
    ####################################################

    prot.insert_mol_sys(mol_gromacs=pep, mol_num=10,
                        new_name='SH3_DD',
                        out_folder=os.path.join(tmp_path, 'top_DD_SH3'))
    ################################
    # ##   Minimize the system   ###
    ################################
    prot.em_2_steps(out_folder=os.path.join(tmp_path, 'em_DD_SH3'),
                    no_constr_nsteps=10, constr_nsteps=10)
    prot_lig_coor = pdb_manip.Coor(prot.coor_file)
    assert 15300 < prot_lig_coor.num < 15518

    prot.display_history()



#!/usr/bin/env python3
# coding: utf-8
##################################
# ######   Free Energy   #########
##################################

import logging
import math
import os
import copy
import sys
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt


from os_command_py import os_command
from pdb_manip_py import pdb_manip

from . import gmx
from .gmx import GROMACS_MOD_DIRNAME
from .tools import monitor, ambertools

# Autorship information
__author__ = "Samuel Murail"
__copyright__ = "Copyright 2020, RPBS"
__credits__ = ["Samuel Murail"]
__license__ = "GNU General Public License v2.0"
__maintainer__ = "Samuel Murail"
__email__ = "samuel.murail@u-paris.fr"
__status__ = "Development"


# Logging
logger = logging.getLogger(__name__)


def show_log():
    """ To use only with Doctest !!!
    Redirect logger output to sys.stdout
    """
    # Delete all handlers
    logger.handlers = []
    # Set the logger level to INFO
    logger.setLevel(logging.INFO)
    # Add sys.sdout as handler
    logger.addHandler(logging.StreamHandler(sys.stdout))

    gmx.show_log()


# Check if Readthedoc is launched skip the program path searching
on_rtd = os.environ.get('READTHEDOCS') == 'True'
if on_rtd:
    logger.info("Gromacs cannot be found")
    GMX_BIN = ""
else:
    GMX_BIN = os_command.which('gmx')


# Boltzmann Constant
KB = 8.31446261815324


class FreeEner:
    """ Free Energy encapsulation class.

    This class can be used to launch and analyze free energy
    calculations using Free Energy Perturbation.

    :param out_folder: output folder
    :type out_folder: str

    :param lambda_coul: lambda points for Coulomb
    :type lambda_coul: list of float

    :param lambda_vdw: lambda points for Lennard Jones
    :type lambda_vdw: list of float

    :param lambda_restr: lambda points for restraints
    :type lambda_restr: list of float

    :param xvg_file_list: List of free energy xvg files
    :type xvg_file_list: list of string

    :param lambda_sys_list: List of lambda GmxSys
    :type lambda_sys_list: list of string

    :param temp: Temperature
    :type temp: float

    :param smile: Ligand SMILE
    :type smile: str

    """

    def __init__(self, mol_name, out_folder, unit='kcal'):
        self.mol_name = mol_name[:3]
        if len(mol_name) > 4:
            logger.warning(
                f'Molecule name {mol_name} is longer than 3 characters,'
                f'name changed to {self.mol_name}')
        self.unit = unit
        self.out_folder = out_folder
        # Lambda
        self.lambda_coul = []
        self.lambda_vdw = []
        self.lambda_restr = []
        # Xvg file list
        self.xvg_file_list = []
        self.lambda_sys_list = []
        self.temp = None
        self.smile = None

    @property
    def conv_fac(self):
        """ Conversion factor from kT to
        self.unit
        """
        return(FreeEner.get_conv_fac(self.unit, self.temp))

    @staticmethod
    def get_conv_fac(unit, temp):
        """ Conversion factor from kT to
        self.unit
        """
        if unit == 'kcal':
            return(0.593)
        elif unit == 'kJ':
            return(4.184 * 0.593)
        elif unit == 'kT':
            return(1)
        elif unit == 'logP':
            return(-1.365679 * KB * temp / 1000)

    @property
    def unit_name(self):
        if self.unit == 'kcal':
            return('kcal/mol')
        elif self.unit == 'kJ':
            return('kJ/mol')
        elif self.unit == 'kT':
            return('kT')
        elif self.unit == 'logP':
            return('logP')

    @property
    def unit_graph(self):
        """ Return unit as math latex for matplotlib
        label purpose.

        """
        if self.unit == 'kcal':
            return(r'kcal \cdot mol^{-1}')
        elif self.unit == 'kJ':
            return(r'kJ \cdot mol^{-1}')
        elif self.unit == 'kT':
            return(r'k_{B}T')
        elif self.unit == 'logP':
            return(r'log \; P')

    def water_box_from_SMILE(self, smile, method_3d='rdkit',
                             iter_num=5000, box_dist=1.1):
        """ Create a water box coordinates and toplogie with a
        molecule defined by its SMILE.

        :param smile: Molecule's SMILE
        :type smile: str

        :param method_3d: Method to compute 3D coordinates
        :type method_3d: str, default='rdkit'

        :param iter_num: Iteration step number for 3D coordinate
            computation
        :type iter_num: int, default=5000

        :param box_dist: Ditance to egde box (nm)
        :type box_dist: float, default=1.1 nm

        .. Note::
            Default box dist 1.1 nm was taken as minimal distance for
            CH4 molecule, to allow domain decomposition with `gmx mdrun`.

        """

        self.smile = smile

        os.makedirs(self.out_folder, exist_ok=True)

        # Create 3d structure from SMILE
        pdb_file = os.path.join(self.out_folder, f'{self.mol_name}.pdb')
        charge = ambertools.smile_to_pdb(smile, pdb_file,
                                         method_3d=method_3d,
                                         mol_name=self.mol_name,
                                         iter_num=iter_num)
        logger.info(f'ligand charge is {charge}')

        # Topologie and system creation
        self.gmxsys = gmx.GmxSys(name=self.mol_name, coor_file=pdb_file)
        self.gmxsys.prepare_top_ligand(
            out_folder=os.path.join(self.out_folder, 'mol_top'),
            ff='amber99sb-ildn', include_mol={self.mol_name: smile})

        self.gmxsys.create_box(name=None, dist=box_dist)
        self.gmxsys.solvate_box(
            out_folder=os.path.join(self.out_folder, 'mol_water_top'))

    def octanol_box_from_SMILE(self, smile, method_3d='rdkit',
                               iter_num=5000, box_dist=1.3):
        """ Create an octanol box coordinates and toplogie with a
        molecule defined by its SMILE.

        :param smile: Molecule's SMILE
        :type smile: str

        :param method_3d: Method to compute 3D coordinates
        :type method_3d: str, default='rdkit'

        :param iter_num: Iteration step number for 3D coordinate
            computation
        :type iter_num: int, default=5000

        :param box_dist: Ditance to egde box (nm)
        :type box_dist: float, default=1.3 nm

        .. Note::
            Default box dist 1.3 nm was taken as minimal distance for
            CH4 molecule, to allow domain decomposition with `gmx mdrun`.
        """

        self.smile = smile

        os.makedirs(self.out_folder, exist_ok=True)

        # Create 3d structure from SMILE
        pdb_file = os.path.join(self.out_folder, f'{self.mol_name}.pdb')
        charge = ambertools.smile_to_pdb(smile, pdb_file,
                                         method_3d=method_3d,
                                         mol_name=self.mol_name,
                                         iter_num=iter_num)
        logger.info(f'ligand charge is {charge}')

        # Topologie and system creation
        self.gmxsys = gmx.GmxSys(name=self.mol_name, coor_file=pdb_file)
        self.gmxsys.prepare_top_ligand(
            out_folder=os.path.join(self.out_folder, 'mol_top'),
            ff='amber99sb-ildn', include_mol={self.mol_name: smile})

        self.gmxsys.create_box(name=None, dist=box_dist)
        self.gmxsys.solvate_box(
            out_folder=os.path.join(self.out_folder, 'mol_octanol_top'),
            cs=os.path.join(GROMACS_MOD_DIRNAME,
                            "template/octanol/prod_OCO.gro"))

        # Add octanol topologie:
        sys_top = gmx.TopSys(self.gmxsys.top_file)
        atomtypes_itp = os.path.join(GROMACS_MOD_DIRNAME,
                                     "template/octanol/OCO_GMX_atomtypes.itp")
        sys_top.add_atomtypes(atomtypes_itp)

        mol_itp = os.path.join(GROMACS_MOD_DIRNAME,
                               "template/octanol/OCO_GMX.itp")
        sys_top.add_mol_itp(mol_itp)

        sys_top.write_file(self.gmxsys.top_file)

    def equilibrate_solvent_box(self, em_steps=10000, dt=0.002, prod_time=10.0,
                                short_steps=50000, temp=300.0):
        """ Equilibrate a solvent (water, octanol) box.

        :param em_steps: Energy minimisation step number
        :type em_steps: int, default=10000

        :param dt: Integration time step (ps)
        :type dt: float, default=0.002 ps

        :param prod_time: Equilibration time (ns)
        :type prod_time: float, default=10.0 ns

        :param short_steps: Short equilibration steps
        :type short_steps: int, default=50000

        :param temp: Temperature of equilibration
        :type temp: float, default=300.0

        """

        # EM
        self.gmxsys.em_2_steps(out_folder=os.path.join(
                                self.out_folder, 'mol_solv_em'),
                               no_constr_nsteps=em_steps,
                               constr_nsteps=em_steps,
                               posres="",
                               emtol=0.1, nstxout=100, maxwarn=3,
                               create_box_flag=False)
        mdp_options = {'tc-grps': 'System',
                       'tau_t': '0.1',
                       'ref_t': temp,
                       }
        # Equilibration
        # first short step with very small dt
        self.gmxsys.production(
            out_folder=os.path.join(self.out_folder, 'mol_solv_prod'),
            name='short_equi',
            nsteps=short_steps, dt=0.0005,
            maxwarn=1, **mdp_options)
        # second step
        prod_steps = 1000 * prod_time / dt
        self.gmxsys.production(
            out_folder=os.path.join(self.out_folder, 'mol_solv_prod'),
            nsteps=prod_steps,
            dt=dt, maxwarn=1, **mdp_options)

    def prepare_complex_pdb(self, pdb_in, smile, ff='amber99sb-ildn'):
        """ Prepare topologie from a pdb file, create box around and solvate it.

        :param pdb_in: Input PDB file
        :type pdb_in: str

        :param smile: Ligand's SMILE
        :type smile: str

        :param ff: Forcefield
        :type ff: str, default='amber99sb-ildn'

        """

        self.gmxsys = gmx.GmxSys(name=self.mol_name, coor_file=pdb_in)
        self.gmxsys.prepare_top(out_folder=os.path.join(
                                    self.out_folder, 'complex_top'),
                                ff=ff, include_mol={self.mol_name: smile})
        self.gmxsys.solvate_add_ions(out_folder=os.path.join(
                                        self.out_folder, 'sys_top'),
                                     ion_C=0.15, maxwarn=3,
                                     create_box_flag=True)

    def equilibrate_complex(self, em_steps=10000, HA_time=0.25,
                            CA_time=0.5, CA_LOW_time=1.0,
                            dt=0.002, dt_HA=0.001, temp=300,
                            receptor_grp='Protein',
                            short_steps=50000):
        """ Equilibrate a receptor-ligand complex system.

        :param em_steps: Energy minimisation step number, default=10000
        :type em_steps: int

        :param HA_time: Heavy atoms restraints equilibration time (ns)
        :type HA_time: float, default=0.25

        :param CA_time: Alpha carbon atoms restraints equilibration
            time (ns)
        :type CA_time: float, default=0.5

        :param CA_LOW_time: Low alpha carbon atoms restraints equilibration
            time (ns)
        :type CA_LOW_time: float, default=1.0

        :param dt: Integration time step (ps)
        :type dt: float, default=0.002 ps

        :param dt_HA: Integration time step (ps)
        :type dt_HA: float, default=0.001 ps

        :param temp: Temperature of equilibration (K)
        :type temp: float, default=300.0 K

        :param receptor_grp: Receptor group (for temperature coupling)
        :type receptor_grp: str, default='Protein'

        :param short_steps: Short equilibration steps
        :type short_steps: int, default=50000

        """

        self.gmxsys.em_2_steps(out_folder=os.path.join(
                                self.out_folder, 'prot_em'),
                               no_constr_nsteps=em_steps,
                               constr_nsteps=em_steps,
                               posres="",
                               maxwarn=3,
                               create_box_flag=False)

        self.gmxsys.convert_trj(select='0' + '\n System', center='yes')

        self.ref_coor = self.gmxsys.coor_file

        # Short equilibration with molecule in its own temp group
        mdp_options = {'nsteps': short_steps,
                       'define': '-DPOSRES', 'dt': 0.0005,
                       'tc-grps': '{} {} Water_and_ions'.format(
                            receptor_grp,
                            self.mol_name),
                       'tau_t': '0.1 0.1 0.1',
                       'ref_t': f'{temp} {temp} {temp}'}

        equi_template_mdp = os.path.join(GROMACS_MOD_DIRNAME,
                                         "template/equi.mdp")

        self.gmxsys.run_md_sim(out_folder=os.path.join(
                                self.out_folder, "sys_short_equi"),
                               name="equi_HA_short_" + self.mol_name,
                               mdp_template=equi_template_mdp,
                               mdp_options=mdp_options, maxwarn=3,
                               pdb_restr=self.ref_coor)

        mdp_options = {}
        if receptor_grp != 'Protein':
            index_dict = self.gmxsys.get_index_dict()
            self.gmxsys.add_ndx('{}|{} \nq\n'.format(
                    index_dict[receptor_grp],
                    index_dict[self.mol_name]),
                check_file_out=False)

            mdp_options = {'tc-grps': '{}_{} Water_and_ions'.format(
                            receptor_grp,
                            self.mol_name)}
        print(mdp_options)

        self.gmxsys.equi_three_step(out_folder=os.path.join(
                                     self.out_folder, 'sys_equi'),
                                    nsteps_HA=1000 * HA_time / dt_HA,
                                    nsteps_CA=1000 * CA_time / dt,
                                    nsteps_CA_LOW=1000 * CA_LOW_time / dt,
                                    dt=dt, dt_HA=dt_HA, maxwarn=3,
                                    pdb_restr=self.ref_coor,
                                    **mdp_options)

    def run(self, lambda_coul_list, lambda_vdw_list, lambda_restr_list=[],
            mbar=False, dir_name='free_ener_run',
            em_steps=5000, nvt_time=10, npt_time=10, prod_time=100,
            dt=0.002, temp=300.0,
            temp_groups='Protein non-Protein', maxwarn=1,
            monitor_tool=monitor.PROGRESS_BAR):
        """Compute free energy to uncouple a molecule to a
        system.

        :param dir_name: path of the output folder
        :type dir_name: str, default='free_ener_run'

        :param mol_name: Name of the molecule
        :type mol_name: str

        :param lambda_coul_list: List lambda points for Coulomb
        :type lambda_coul_list: list

        :param lambda_vdw_list: List lambda points for Lennard Jones
        :type lambda_vdw_list: list

        :param lambda_bond_list: List lambda points for restraints
        :type lambda_bond_list: list, default=[]

        :param mbar: MBAR flag
        :type mbar: bool, default=False

        :param em_steps: number of minimisation steps
        :type em_steps: int, default=5000

        :param nvt_time: Time (ps) of NVT equilibration
        :type nvt_time: int, default=10 ps

        :param npt_time:  Time (ps) of NPT equilibration
        :type npt_time: int, default=10 ps

        :param prod_time: Time (ps) of production run
        :type prod_time: float, default=100 ps

        :param dt: integration time step
        :type dt: float, default=0.002

        :param name: name of the simulation to run
        :type name: str, default=None

        :param temp: Temperature K
        :type temp: float, default=300.0

        :param temp_groups: Group(s) for temperature coupling
        :type temp_groups: str, default='Protein non-Protein'

        :param maxwarn: Maximum number of warnings when using ``gmx grompp``
        :type maxwarn: int, default=0

        :param monitor: option to monitor a simulation, if not none monitor
            should contains two values: ``function`` the function to be ran
            while simulation is running and ``input`` parameters for the
            function
        :type rerun: dict, default=None

        **Object requirement(s):**

            * self.gmxsys.coor_file
            * self.gmxsys.top_file
            * self.gmxsys.nt
            * self.gmxsys.ntmpi
            * self.gmxsys.gpu_id

        **Object field(s) changed:**

            * self.lambda_sys_list
            * self.lambda_coul
            * self.lambda_vdw
            * self.lambda_restr
            * self.prod_time
            * self.xvg_file_list

        """

        if monitor.isnotebook():
            from tqdm.notebook import tqdm
        else:
            from tqdm import tqdm

        # Remove lambda=0 for vdw and elec
        # If previous restr lambdas are computed
        if len(lambda_restr_list) > 0:
            lambda_coul_list = list(lambda_coul_list)
            lambda_coul_list.remove(0.0)
        if len(lambda_restr_list) + len(lambda_coul_list) > 0:
            lambda_vdw_list = list(lambda_vdw_list)
            lambda_vdw_list.remove(0.0)

        lambda_restr_num = len(lambda_restr_list)
        lambda_coul_num = len(lambda_coul_list)
        lambda_vdw_num = len(lambda_vdw_list)

        restr_lambdas = "".join([
            '{:.2f} '.format(i) for i in lambda_restr_list]) +\
            "1.00 " * (lambda_vdw_num + lambda_coul_num)
        coul_lambdas = "0.00 " * lambda_restr_num + "".join([
            '{:.2f} '.format(i) for i in lambda_coul_list]) +\
            "1.00 " * (lambda_vdw_num)
        vdw_lambdas = "0.00 " * (lambda_coul_num + lambda_restr_num) +\
            "".join(['{:.2f} '.format(i) for i in lambda_vdw_list])

        logger.info(f"Coulomb lambda : {coul_lambdas}\n"
                    f"Vdw lambda     : {vdw_lambdas}\n"
                    f"restr_lambdas  : {restr_lambdas}")

        # https://events.prace-ri.eu/event/674/attachments/618/896/MD_FreeEnergyTutorial.pdf
        # Suggest to use tau_t = 1.0 to avoid
        # over-damping the dynamics of water
        tau_t_str = " ".join(['{}'.format(1.0)
                              for _ in range(len(temp_groups.split()))])
        ref_t_str = " ".join(['{}'.format(temp)
                              for _ in range(len(temp_groups.split()))])

        free_ener_option_md = {'integrator': 'sd',
                               'dt': dt,
                               'constraints': 'all-bonds',
                               'nstcalcenergy': 50,
                               'tcoupl': '',
                               'tc_grps': temp_groups,
                               'tau_t': tau_t_str,
                               'ref_t': ref_t_str,
                               'vdwtype': 'cut_off',
                               'vdw_modifier': 'force_switch',
                               'rvdw_switch': 1.0,
                               'rvdw': 1.1,
                               'coulombtype': 'pme',
                               'rcoulomb': 1.1,
                               'lincs_order': 8,  # Try to fix the LINCS Errors
                               'free_energy': 'yes',
                               'init_lambda-state': 0,
                               'calc-lambda-neighbors': 1,
                               'delta_lambda': 0,
                               'coul_lambdas': coul_lambdas,
                               'vdw_lambdas': vdw_lambdas,
                               'bonded_lambdas': restr_lambdas,
                               'sc_alpha': 0.5,
                               'sc_power': 1,
                               'sc_sigma': 0.3,
                               'couple_moltype': self.mol_name,
                               'couple_lambda0': 'vdw-q',
                               'couple_lambda1': 'none',
                               'couple_intramol': 'no',
                               'nstdhdl': 50,
                               'separate_dhdl_file': 'yes'}

        self.xvg_file_list = []

        nvt_steps = int(nvt_time / dt)
        npt_steps = int(npt_time / dt)
        prod_steps = int(prod_time / dt)

        tot_step = (lambda_restr_num + lambda_coul_num + lambda_vdw_num) * (
            em_steps + nvt_steps + npt_steps + prod_steps)
        pbar = tqdm(total=tot_step)

        for i in range(lambda_restr_num + lambda_coul_num + lambda_vdw_num):

            logger.info('Compute lambda {} / {}'.format(
                i + 1, lambda_restr_num + lambda_coul_num + lambda_vdw_num))

            lambda_sys = FreeEner.compute_lambda_point(
                self.gmxsys, i,
                self.mol_name,
                os.path.join(self.out_folder,
                             dir_name),
                free_ener_option_md, pbar,
                mbar=mbar,
                em_steps=em_steps,
                nvt_steps=nvt_steps,
                npt_steps=npt_steps,
                prod_steps=prod_steps,
                maxwarn=maxwarn,
                monitor_tool=monitor_tool)
            self.lambda_sys_list.append(lambda_sys)
            self.xvg_file_list.append(lambda_sys.xvg)

        self.temp = temp
        self.lambda_coul = list(lambda_coul_list)
        self.lambda_vdw = list(lambda_vdw_list)
        self.lambda_restr = list(lambda_restr_list)
        self.prod_time = prod_time

    @staticmethod
    def compute_lambda_point(gmx_sys, lambda_iter, mol_name, out_folder,
                             free_ener_option, pbar, mbar,
                             em_steps, nvt_steps, npt_steps,
                             prod_steps, maxwarn=1,
                             monitor_tool=monitor.PROGRESS_BAR):
        """ Run the different steps of a single lambda point.

        :param gmx_sys: Gmx System to start from
        :type gmx_sys: GmxSys

        :param lambda_iter: lambda point
        :type lambda_iter: int

        :param mol_name: Molecule residue name
        :type mol_name: str

        :param out_folder: path of the output folder
        :type out_folder: str

        :param free_ener_option: Mdp options
        :type free_ener_option: dict

        :param pbar: progress bar object
        :type pbar: tqmd bar

        :param mbar: MBAR flag
        :type mbar: bool

        :param em_steps: number of minimisation steps
        :type em_steps: int

        :param nvt_steps: number of NVT equilibration steps
        :type nvt_steps: int

        :param npt_steps:  number of NPT equilibration steps
        :type npt_steps: int

        :param prod_steps: number of production steps
        :type prod_steps: int

        :param maxwarn: Maximum number of warnings when using ``gmx grompp``
        :type maxwarn: int, default=0

        :param monitor: option to monitor a simulation, if not none monitor
            should contains two values: ``function`` the function to be ran
            while simulation is running and ``input`` parameters for the
            function
        :type rerun: dict, default=None

        """

        start_sys = copy.deepcopy(gmx_sys)

        mini_template_mdp = os.path.join(
            GROMACS_MOD_DIRNAME, "template/mini.mdp")
        equi_template_mdp = os.path.join(
            GROMACS_MOD_DIRNAME, "template/equi.mdp")

        free_ener_option_md = copy.deepcopy(free_ener_option)
        free_ener_option_em = copy.deepcopy(free_ener_option)

        # Remove useless options for EM
        del(free_ener_option_em['integrator'],
            free_ener_option_em['dt'],
            free_ener_option_em['tc_grps'],
            free_ener_option_em['tau_t'],
            free_ener_option_em['ref_t'])

        sys_name = '{}_vdwq_{:02d}'.format(mol_name, lambda_iter)

        free_ener_option_em.update({'init_lambda-state': lambda_iter,
                                    'nsteps': em_steps})

        # Mini 5000 steps
        mol_sys = copy.deepcopy(start_sys)
        mol_sys.run_md_sim(out_folder=os.path.join(
                            out_folder, '00_em'),
                           name='em_' + sys_name,
                           mdp_template=mini_template_mdp,
                           mdp_options=free_ener_option_em, maxwarn=maxwarn,
                           monitor_tool=monitor_tool)
        pbar.update(em_steps)

        # MD
        free_ener_option_md.update({'init_lambda-state': lambda_iter})

        # NVT equilibration
        free_ener_option_md.update({'nsteps': nvt_steps,
                                    'pcoupl': 'no'})
        mol_sys.run_md_sim(out_folder=os.path.join(out_folder,
                                                   '01_equi_nvt'),
                           name='nvt_' + sys_name,
                           mdp_template=equi_template_mdp,
                           mdp_options=free_ener_option_md, maxwarn=maxwarn,
                           monitor_tool=monitor_tool)
        pbar.update(nvt_steps)

        # NPT equilibration
        free_ener_option_md.update({'nsteps': npt_steps,
                                    'pcoupl': 'parrinello-Rahman'})
        mol_sys.run_md_sim(out_folder=os.path.join(out_folder,
                                                   '02_equi_npt'),
                           name='npt_' + sys_name,
                           mdp_template=equi_template_mdp,
                           mdp_options=free_ener_option_md, maxwarn=maxwarn,
                           monitor_tool=monitor_tool)
        pbar.update(npt_steps)

        # Production
        if mbar:
            free_ener_option_md.update({'calc-lambda-neighbors': -1})

        free_ener_option_md.update({'nsteps': prod_steps})
        mol_sys.run_md_sim(out_folder=os.path.join(out_folder, '03_prod'),
                           name='prod_' + sys_name,
                           mdp_template=equi_template_mdp,
                           mdp_options=free_ener_option_md, maxwarn=maxwarn,
                           monitor_tool=monitor_tool)
        pbar.update(prod_steps)

        return(mol_sys)

    def extend_lambda_prod(self, prod_time):
        """ Extend free energy production.

        :param prod_time: production time (ns)
        :type prod_time: float

        """

        dt = float(self.lambda_sys_list[0].get_mdp_dict()['dt'])
        nsteps = prod_time / dt
        lambda_num = len(self.lambda_restr) + len(self.lambda_coul) +\
            len(self.lambda_vdw)
        xvg_file_list = []

        for i in range(lambda_num):
            logger.info(f'Compute lambda {i+1} / {lambda_num}')

            self.lambda_sys_list[i].extend_sim(nsteps=nsteps)
            output = self.lambda_sys_list[i].get_all_output()
            xvg_file_list += output['xvg']

        self.prod_time = prod_time
        self.xvg_file_list = xvg_file_list

    def align_ref_traj(self, rec_group='Protein'):
        """
        """

        # Center and align traj:
        self.gmxsys.convert_trj(
            select=rec_group + '\n System', center='yes')
        self.gmxsys.convert_trj(
            select=rec_group + '\n System', fit='rot+trans', pbc='none')

    def compute_add_intermol_from_traj(self, ref_coor=None,
                                       rec_group='Protein', k=41.84,
                                       cutoff_prot=6.0):
        """ Compute intermolecular restraint from the last GmxSys trajectory.
        Get a coor object
        Get distance and angles
        """

        if ref_coor is None:
            ref_coor = self.ref_coor

        self.align_ref_traj(rec_group=rec_group)

        lig_atom_list = self.get_ligand_atoms(ref_coor)

        rec_atom_list = self.get_protein_atoms_from_rmsf(
            ref_coor,
            lig_atom_list,
            rec_group=rec_group,
            cutoff_max=cutoff_prot)

        self.add_intermol_restr_index(rec_atom_list, lig_atom_list,
                                      ref_coor, k=k)

    def get_protein_atoms_from_rmsf(self, ref_coor, lig_atom_list,
                                    rec_group='Protein',
                                    cutoff_max=6.0):
        """
        """
        coor = pdb_manip.Coor(ref_coor)

        # Get protein RMSF
        prot_rmsf = self.gmxsys.get_rmsf([rec_group])

        # Get backbone protein atom around the ligand:
        if rec_group == 'Protein':
            bk_atoms_list = ['C', 'CA', 'N', 'O']
        elif rec_group == 'DNA':
            bk_atoms_list = ['C4\'', 'C3\'', 'O3\'', 'O5\'', 'P', 'C5\'']
        backbone = coor.select_part_dict({'name': bk_atoms_list})
        # Create coor for the 3 ligand atoms
        atom_coor = pdb_manip.Coor()
        atom_coor.atom_dict = {
            i: coor.atom_dict[lig_atom_list[i] - 1] for i in range(3)}
        around_atom = backbone.get_index_dist_between(
            atom_coor, cutoff_min=0.0, cutoff_max=cutoff_max)
        # Extract RMSF of contact atoms
        around_lig_df = prot_rmsf[prot_rmsf.Atom.isin(around_atom)]
        around_lig_df = around_lig_df.sort_values(by=['RMSF'])
        # Get residue of stablest atom:
        uniq_res = coor.atom_dict[
            around_lig_df.reset_index().loc[0, 'Atom']]['uniq_resid']
        rec_atom_list = coor.get_index_selection(
            {'uniq_resid': [uniq_res], 'name': [bk_atoms_list[0]]})
        rec_atom_list += coor.get_index_selection(
            {'uniq_resid': [uniq_res], 'name': [bk_atoms_list[1]]})
        rec_atom_list += coor.get_index_selection(
            {'uniq_resid': [uniq_res], 'name': [bk_atoms_list[2]]})

        # Atom index need to be +1 to be in gromacs numbering
        rec_atom_list = [index + 1 for index in rec_atom_list]
        logger.debug(f'Receptor atom indexes : {rec_atom_list}')

        return(rec_atom_list)

    def get_protein_atoms_from_res(self, resid,
                                   rec_group='Protein'):
        """
        """

        coor = pdb_manip.Coor(self.ref_coor)

        # Get backbone protein atom around the ligand:
        if rec_group == 'Protein':
            bk_atoms_list = ['C', 'CA', 'N']
        elif rec_group == 'DNA':
            bk_atoms_list = ['C4\'', 'C3\'', 'O3\'']

        rec_atom_list = coor.get_index_selection(
            {'res_num': [resid], 'name': [bk_atoms_list[0]]})
        rec_atom_list += coor.get_index_selection(
            {'res_num': [resid], 'name': [bk_atoms_list[1]]})
        rec_atom_list += coor.get_index_selection(
            {'res_num': [resid], 'name': [bk_atoms_list[2]]})

        # Atom index need to be +1 to be in gromacs numbering
        rec_atom_list = [index + 1 for index in rec_atom_list]
        logger.debug(f'Receptor atom indexes : {rec_atom_list}')

        return(rec_atom_list)

    def get_ligand_atoms(self, ref_coor):
        """
        """

        coor = pdb_manip.Coor(ref_coor)
        # Get RMSF for ligand:
        lig_rmsf = self.gmxsys.get_rmsf([self.mol_name])
        # Sort df by RMSF
        lig_rmsf = lig_rmsf.sort_values(by=['RMSF'])
        # Get the stablest atom which is not Hydrogen
        for i, row in lig_rmsf.iterrows():
            atom = coor.atom_dict[int(row['Atom']) - 1]
            if not atom['name'].startswith('H'):
                lig_atom_list = [int(row['Atom']) - 1]
                break
        # Get connected atoms:
        lig_coor = coor.select_part_dict({'res_name': [self.mol_name]})
        atom_coor = pdb_manip.Coor()

        # Add 2 close atom consecutively
        cutoff_max = 1.5
        while len(lig_atom_list) < 3:
            cutoff_max += 0.1
            atom_coor.atom_dict = {0: coor.atom_dict[lig_atom_list[-1]]}
            close_lig_atoms = lig_coor.get_index_dist_between(
                atom_coor, cutoff_min=0.1, cutoff_max=cutoff_max)
            if len(close_lig_atoms) > 0:
                for _, row in lig_rmsf.iterrows():
                    num = int(row['Atom']) - 1
                    atom = coor.atom_dict[num]
                    if num in close_lig_atoms and not (
                            atom['name'].startswith('H')
                            or num in lig_atom_list):
                        lig_atom_list.append(num)
                        break

        # Atom index need to be +1 to be in gromacs numbering
        lig_atom_list = [index + 1 for index in lig_atom_list]
        logger.debug(f'Ligand atom indexes : {lig_atom_list}')
        print(f'Ligand atom indexes : {lig_atom_list}')

        # In rare case there is more than 3 atoms
        return(lig_atom_list[:3])

    def add_intermol_restr_index(self, rec_index_list, lig_index_list,
                                 ref_coor, k=41.84, temp=300):
        r"""Compute and add the intermolecular restraints based on the
        ref_coor distance and angles.

        Give three atoms for each receptor and ligand index list:
        :math:`R_0`, :math:`R_1`, :math:`R_2` and :math:`L_0`, :math:`L_1`,
        :math:`L_2`
        Will define:

        * 1 bond:

            * :math:`R_0 - L_0`

        * 2 angles:

            * :math:`R_0 - L_0 - L_1`
            * :math:`R_1 - R_0 - L_0`

        * 2 dihedral angles:

            * :math:`R_0 - L_0 - L_1 - L_2`
            * :math:`R_2 - R_1 - R_0 - L_0`

        :param rec_index_list: List of the receptor atom index
        :type rec_index_list: list

        :param lig_index_list: List of the ligand atom index
        :type lig_index_list: list

        :param ref_coor: Reference coordinates file
        :type ref_coor: str

        :param k: intermolecular force constant,
            (default= :math:`41.84 \; kcal \, mol^{-1} \, nm^{-2}`)
        :type k: float

        :param temp: Temperature defalult=300 K
        :type temp: float

        **Object requirement(s):**

            * self.gmxsys.coor_file
            * self.gmxsys.top_file

        **Object field(s) changed:**

            * self.gmxsys.top_file

        """
        bond_type = 6
        angle_type = 1
        dihed_type = 2

        coor = pdb_manip.Coor(ref_coor)

        # R0-L0
        # Index from gromacs starts at 1 and in pdb_manip at 0
        dist = pdb_manip.Coor.atom_dist(
            coor.atom_dict[rec_index_list[0] - 1],
            coor.atom_dict[lig_index_list[0] - 1]) / 10.0
        bond_list = [[rec_index_list[0],
                      lig_index_list[0],
                      bond_type,
                      round(dist, 3),
                      0,
                      round(dist, 3),
                      k * 100]]

        # R0-L0-L1
        angle_1 = pdb_manip.Coor.atom_angle(
            coor.atom_dict[rec_index_list[0] - 1],
            coor.atom_dict[lig_index_list[0] - 1],
            coor.atom_dict[lig_index_list[1] - 1])
        # R1-R0-L0
        angle_2 = pdb_manip.Coor.atom_angle(
            coor.atom_dict[rec_index_list[1] - 1],
            coor.atom_dict[rec_index_list[0] - 1],
            coor.atom_dict[lig_index_list[0] - 1])
        angle_list = [[rec_index_list[0],
                       lig_index_list[0],
                       lig_index_list[1],
                       angle_type,
                       round(angle_1, 3), 0,
                       round(angle_1, 3), k]]
        angle_list += [[rec_index_list[1],
                        rec_index_list[0],
                        lig_index_list[0],
                        angle_type,
                        round(angle_2, 3), 0,
                        round(angle_2, 3), k]]

        # R0-L0-L1-L2
        dihed_1 = pdb_manip.Coor.atom_dihed_angle(
            coor.atom_dict[rec_index_list[0] - 1],
            coor.atom_dict[lig_index_list[0] - 1],
            coor.atom_dict[lig_index_list[1] - 1],
            coor.atom_dict[lig_index_list[2] - 1])
        # R2-R1-R0-L0
        dihed_2 = pdb_manip.Coor.atom_dihed_angle(
            coor.atom_dict[rec_index_list[2] - 1],
            coor.atom_dict[rec_index_list[1] - 1],
            coor.atom_dict[rec_index_list[0] - 1],
            coor.atom_dict[lig_index_list[0] - 1])
        # R1-R0-L0-L1
        dihed_3 = pdb_manip.Coor.atom_dihed_angle(
            coor.atom_dict[rec_index_list[1] - 1],
            coor.atom_dict[rec_index_list[0] - 1],
            coor.atom_dict[lig_index_list[0] - 1],
            coor.atom_dict[lig_index_list[1] - 1])
        dihed_list = [[rec_index_list[0],
                       lig_index_list[0],
                       lig_index_list[1],
                       lig_index_list[2],
                       dihed_type,
                       round(dihed_1, 3), 0,
                       round(dihed_1, 3), k]]
        dihed_list += [[rec_index_list[2],
                        rec_index_list[1],
                        rec_index_list[0],
                        lig_index_list[0],
                        dihed_type,
                        round(dihed_2, 3), 0,
                        round(dihed_2, 3), k]]

        dihed_list += [[rec_index_list[1],
                        rec_index_list[0],
                        lig_index_list[0],
                        lig_index_list[1],
                        dihed_type,
                        round(dihed_3, 3), 0,
                        round(dihed_3, 3), k]]

        top = gmx.TopSys(self.gmxsys.top_file)

        top.add_intermolecular_restr(bond_list=bond_list,
                                     angle_list=angle_list,
                                     dihed_list=dihed_list)

        top.write_file(self.gmxsys.top_file[:-4] + '_rest.top')
        self.gmxsys.top_file = self.gmxsys.top_file[:-4] + '_rest.top'

    def get_water_restr(self, temp=300):
        r""" Compute ligand restaint energy in water using
        Boresh et al. equation:

        :math:`\Delta G_{restr} = RT \ln \left( \dfrac{8 \pi^2 V^0}
        {r^2_0 \sin{\theta_{a 0}} \sin{\theta_{b 0}}} \dfrac{
        \sqrt{k_r k_{\theta_a} k_{\theta_b} k_{\tau_\alpha} k_{\tau_\beta}
        k_{\tau_\gamma}}} {2 \pi KT^3} \right)`

        """

        top = gmx.TopSys(self.gmxsys.top_file)

        dist = float(top.inter_bond_list[0]['rA'])
        k_bond = float(top.inter_bond_list[0]['kB'])

        k_angle = float(top.inter_angl_list[0]['kB'])
        angle_1 = float(top.inter_angl_list[0]['thA'])
        angle_2 = float(top.inter_angl_list[1]['thA'])

        # Compute DG restr in water:
        V0 = 1660 * 1e-3  # nm³
        GK = KB * 1e-3 / 4.184  # Gas constant kcal/mol/K
        # K need to converted to Kcal (currently in KJ)
        k_bond /= 4.184  # k*1e2 kcal/(nm*mol)
        k_angle /= 4.184

        numerator = 8 * math.pi**2 * V0 * (k_bond * k_angle**5)**0.5
        denominator_1 = (dist**2 * math.sin(math.radians(angle_1))
                         * math.sin(math.radians(angle_2)))
        denominator_2 = (2 * math.pi * GK * temp)**3

        dg_rest_water = GK * temp * math.log(
            numerator / (denominator_1 * denominator_2))

        # Convert to kT
        dg_rest_water /= 0.593
        # Convert to self.unit
        dg_rest_water *= self.conv_fac

        return(dg_rest_water)

    def show_intermol_restr(self):
        """ Show traj with atom implied in intermolecular
        restraints using nglview library.
        """

        view = self.gmxsys.view_traj()
        view.clear_representations()
        view.add_representation("cartoon")

        # Get distance index
        top = gmx.TopSys(self.gmxsys.top_file)

        atom_list = []
        for bond in top.inter_bond_list:
            atom_a = int(bond['ai']) - 1
            atom_b = int(bond['aj']) - 1
            view.add_representation("distance",
                                    atomPair=[(f'@{atom_a}',
                                               f'@{atom_b}')])
            atom_list += [atom_a, atom_b]

        # Get all atoms:
        for angle in top.inter_dihe_list:
            atom_a = int(angle['ai']) - 1
            atom_b = int(angle['aj']) - 1
            atom_c = int(angle['ak']) - 1
            atom_d = int(angle['al']) - 1
            atom_list += [atom_a, atom_b, atom_c, atom_d]
        # Get unique value:
        atom_list = list(set(atom_list))

        # Get residue number:
        coor = pdb_manip.Coor(self.ref_coor)
        residue_list = []
        for atom in atom_list:
            residue_list.append(coor.atom_dict[atom]['res_num'])
        residue_list = list(set(residue_list))

        selection = f'[{self.mol_name}]'
        if len(residue_list) > 0:
            selection += ' or '
            selection += ','.join([str(i) for i in residue_list])
        print('residue list:', residue_list)

        view.add_representation("licorice",
                                selection=selection)
        atom_list_sel = '@' + ','.join([str(i) for i in atom_list])
        view.add_representation("ball+stick",
                                selection=atom_list_sel,
                                aspectRatio=4.0)

        return(view)

    def plot_intermol_restr(self, graph_out=None):
        """
        """
        fig, axs = plt.subplots(3, figsize=(8, 10))
        fig.suptitle(r'Inter Molecular Retraints')

        width_ratio = 0.7

        top = gmx.TopSys(self.gmxsys.top_file)

        dist_index_list = []
        for bond in top.inter_bond_list:
            dist = float(bond['rA'])
            dist_index_list.append([int(bond['ai']), int(bond['aj'])])
            axs[0].axvline(x=dist, c='0.5', linestyle='--')

        bond_df = self.gmxsys.get_dist(dist_index_list)

        min_dist = round(bond_df.iloc[:, 1].min(), 2)
        max_dist = round(bond_df.iloc[:, 1].max(), 2)
        print(min_dist, max_dist)
        # bins size of 0.1 Å or 0.01 nm
        bins_num = int((max_dist - min_dist) / 0.01)

        bond_df.iloc[:, 1:].plot.kde(ax=axs[0])
        bond_df.iloc[:, 1:].plot.hist(density=True, ax=axs[0], legend=False,
                                      bins=bins_num,
                                      range=(min_dist, max_dist))
        axs[0].legend(bbox_to_anchor=(1.05, 1), loc='upper left')
        axs[0].set_xlabel('distance ($nm$)')
        axs[0].xaxis.set_label_coords(0.5, -0.1)
        axs[0].grid()
        # To avoid legend being cut:
        box = axs[0].get_position()
        axs[0].set_position([box.x0, box.y0,
                             box.width * width_ratio,
                             box.height])

        angle_index_list = []
        for angle in top.inter_angl_list:
            angle_val = float(angle['thA'])
            angle_index_list.append([int(angle['ai']),
                                     int(angle['aj']),
                                     int(angle['ak'])])
            axs[1].axvline(x=angle_val, c='0.5', linestyle='--')

        angle_df = self.gmxsys.get_angle(angle_index_list)
        angle_df.iloc[:, 1:].plot.kde(ax=axs[1])
        angle_df.iloc[:, 1:].plot.hist(density=True, ax=axs[1],
                                       legend=False, alpha=0.5,
                                       bins=18 * 2, range=(0, 180))
        axs[1].legend(bbox_to_anchor=(1.05, 1), loc='upper left')
        axs[1].set_xlabel(r'angle ($^{\circ}$)')
        axs[1].xaxis.set_label_coords(0.5, -0.1)
        axs[1].set_xlim(0, 180)
        axs[1].grid()
        # To avoid legend being cut:
        box = axs[1].get_position()
        axs[1].set_position([box.x0, box.y0,
                             box.width * width_ratio,
                             box.height])

        dihe_index_list = []
        for angle in top.inter_dihe_list:
            dihe_val = float(angle['thA'])
            dihe_index_list.append([int(angle['ai']),
                                    int(angle['aj']),
                                    int(angle['ak']),
                                    int(angle['al'])])
            axs[2].axvline(x=dihe_val, c='0.5', linestyle='--')

        dihe_df = self.gmxsys.get_angle(dihe_index_list)
        dihe_df.iloc[:, 1:].plot.kde(ax=axs[2])
        dihe_df.iloc[:, 1:].plot.hist(density=True, ax=axs[2],
                                      legend=False, alpha=0.5,
                                      bins=36 * 2, range=(-180, 180))
        axs[2].legend(bbox_to_anchor=(1.05, 1), loc='upper left')
        axs[2].set_xlabel(r'dihedral angle ($^{\circ}$)')
        axs[2].xaxis.set_label_coords(0.5, -0.1)
        axs[2].set_xlim(-180, 180)
        axs[2].grid()
        # To avoid legend being cut:
        box = axs[2].get_position()
        axs[2].set_position([box.x0, box.y0,
                             box.width * width_ratio,
                             box.height])

        if graph_out:
            plt.savefig(graph_out)

    @staticmethod
    def get_bar(xvg_file_list, bar_xvg='bar.xvg',
                barint_xvg='barint.xvg', hist_xvg='histogram.xvg',
                begin_time=0, end_time=-1,
                check_file_out=True, keep_ener_file=False):
        """Get energy of a system using ``gmx bar``.

        I don't know how to compute std like in gmx bar.
        """

        logger.info("- Extract bar energy")

        cmd_convert = os_command.Command([GMX_BIN, "bar",
                                          "-f", *xvg_file_list,
                                          "-o", bar_xvg,
                                          "-oi", barint_xvg,
                                          "-oh", hist_xvg,
                                          "-b", str(begin_time),
                                          "-e", str(end_time)])

        cmd_convert.display()
        cmd_convert.run()

        ener_pd = monitor.read_xvg(bar_xvg)

        if not keep_ener_file:
            os_command.delete_file(bar_xvg)
            os_command.delete_file(barint_xvg)
            os_command.delete_file(hist_xvg)

        return ener_pd

    def get_free_ener(self, begin_time=0, end_time=-1, unit=None):
        """ Show free energy calculation output

        NEED TO FIX STD !!
        """
        if unit is None:
            conv_fac = self.conv_fac
        else:
            conv_fac = FreeEner.get_conv_fac(unit, self.temp)

        self.table = FreeEner.get_bar(self.xvg_file_list,
                                      begin_time=begin_time,
                                      end_time=end_time)

        tot_ener = self.table['DG (kT)'].sum() * conv_fac
        tot_ener_std = sum(self.table['+/-']**2)**0.5 * conv_fac

        self.ener = tot_ener
        self.ener_std = tot_ener_std

        logger.info('\nDDG Tot   = {:5.2f} +/- {:.2f} {}\n'.format(
            tot_ener, tot_ener_std, self.unit_name))

        lambda_bond_num = len(self.lambda_restr)
        lambda_coul_num = len(self.lambda_coul)

        if self.lambda_restr:
            bond_contrib = self.table['DG (kT)'][
                :lambda_bond_num].sum() * conv_fac
            bond_contrib_std = sum(
                self.table['+/-'][:lambda_bond_num]**2)**0.5 * conv_fac

            logger.info('DG Restr = {:5.2f} +/- {:.2f} {}'.format(
                bond_contrib, bond_contrib_std, self.unit_name))

        if self.lambda_coul:
            coulomb_contrib = self.table['DG (kT)'][
                lambda_bond_num:lambda_bond_num + lambda_coul_num].sum()
            coulomb_contrib *= conv_fac
            coulomb_contrib_std = sum(self.table['+/-'][
                lambda_bond_num:lambda_bond_num + lambda_coul_num]**2)**0.5
            coulomb_contrib_std *= conv_fac
            logger.info('DG Coul  = {:5.2f} +/- {:.2f} {}'.format(
                coulomb_contrib, coulomb_contrib_std, self.unit_name))

        if self.lambda_vdw:
            vdw_contrib = self.table['DG (kT)'][
                lambda_bond_num + lambda_coul_num:].sum() * conv_fac
            vdw_contrib_std = sum(
                self.table['+/-'][lambda_bond_num + lambda_coul_num:]**2
                )**0.5 * conv_fac
            logger.info('DG LJ    = {:5.2f} +/- {:.2f} {}'.format(
                vdw_contrib, vdw_contrib_std, self.unit_name))

        return(tot_ener, tot_ener_std)

    def plot_convergence(self, graph_out=None, dt=10):
        """
        """
        try:
            self.compute_convergence_alchemlyb(dt=dt)
        except ImportError:
            self.compute_convergence_gbar(dt=dt)

        return(self.plot_convergence_graph(graph_out=graph_out))

    def compute_convergence_alchemlyb(self, dt=10):
        """
        """

        from alchemlyb.parsing.gmx import extract_dHdl
        from alchemlyb.estimators import TI

        files = self.xvg_file_list
        # NEED TO SORTS FILES IN CASE OF RESTARTS
        restr_num = len(self.lambda_restr)
        coul_num = len(self.lambda_coul)
        vdw_num = len(self.lambda_vdw)

        dHdl_tot = pd.concat([extract_dHdl(xvg, T=self.temp) for xvg in files])
        dHdl_tot_no_index = dHdl_tot.reset_index()
        # Remove duplicates due to restarts
        dHdl_tot_no_index = dHdl_tot_no_index.drop_duplicates(
            subset=['bonded-lambda', 'coul-lambda', 'vdw-lambda', 'time'],
            keep='last')
        # Sort values
        dHdl_tot_no_index = dHdl_tot_no_index.sort_values(
            by=['bonded-lambda', 'coul-lambda', 'vdw-lambda', 'time'])
        # Reset index again:
        dHdl_tot_no_index = dHdl_tot_no_index.reset_index(drop=True)
        # return(dHdl_tot_no_index)

        ti_tot_list = []
        ti_tot_sd_list = []
        time_list = np.arange(dt, self.prod_time + dt, dt)
        index_colnames = ['time', 'coul-lambda', 'vdw-lambda', 'bonded-lambda']
        self.convergence_data = {'time': time_list}

        sim_num_val = len(dHdl_tot_no_index) / (restr_num + coul_num + vdw_num)

        if self.lambda_coul:
            dHdl_coul_no_index = dHdl_tot_no_index[
                max(0, int(sim_num_val * (restr_num - 1))):
                int(sim_num_val * (restr_num + coul_num))]
            ti_coul_list = []
            ti_coul_sd_list = []

        if self.lambda_vdw:
            dHdl_vdw_no_index = dHdl_tot_no_index[
                int(sim_num_val * (restr_num + coul_num - 1)):]
            ti_vdw_list = []
            ti_vdw_sd_list = []

        if self.lambda_restr:
            dHdl_restr_no_index = dHdl_tot_no_index[
                :int(sim_num_val * restr_num)]
            ti_restr_list = []
            ti_restr_sd_list = []

        for time in time_list:
            # Tot
            dHdl_local = dHdl_tot_no_index[
                (dHdl_tot_no_index.time < time) &
                (dHdl_tot_no_index.time >= time - dt)]
            dHdl_local = dHdl_local.set_index(index_colnames)
            ti = TI().fit(dHdl_local)
            ti_tot_list.append(-self.conv_fac * ti.delta_f_.iloc[0, -1])
            ti_tot_sd_list.append(self.conv_fac * ti.d_delta_f_.iloc[0, -1])

            # Restr
            if self.lambda_restr:
                dHdl_local = dHdl_restr_no_index[
                    (dHdl_restr_no_index.time < time) &
                    (dHdl_restr_no_index.time >= time - dt)]
                dHdl_local = dHdl_local.set_index(index_colnames)
                ti = TI().fit(dHdl_local)
                ti_restr_list.append(
                    -1 * self.conv_fac * ti.delta_f_.iloc[0, -1])
                ti_restr_sd_list.append(
                    self.conv_fac * ti.d_delta_f_.iloc[0, -1])

            # Coul
            if self.lambda_coul:
                dHdl_local = dHdl_coul_no_index[
                    (dHdl_coul_no_index.time < time) &
                    (dHdl_coul_no_index.time >= time - dt)]
                dHdl_local = dHdl_local.set_index(index_colnames)
                ti = TI().fit(dHdl_local)
                ti_coul_list.append(-self.conv_fac * ti.delta_f_.iloc[0, -1])
                ti_coul_sd_list.append(
                    self.conv_fac * ti.d_delta_f_.iloc[0, -1])

            # VDW
            if self.lambda_vdw:
                dHdl_local = dHdl_vdw_no_index[
                    (dHdl_vdw_no_index.time < time) &
                    (dHdl_vdw_no_index.time >= time - dt)]
                dHdl_local = dHdl_local.set_index(index_colnames)
                ti = TI().fit(dHdl_local)
                ti_vdw_list.append(-self.conv_fac * ti.delta_f_.iloc[0, -1])
                ti_vdw_sd_list.append(
                    self.conv_fac * ti.d_delta_f_.iloc[0, -1])

        self.convergence_data['tot'] = [ti_tot_list, ti_tot_sd_list]
        if self.lambda_restr:
            self.convergence_data['restr'] = [ti_restr_list, ti_restr_sd_list]
        if self.lambda_coul:
            self.convergence_data['coul'] = [ti_coul_list, ti_coul_sd_list]
        if self.lambda_vdw:
            self.convergence_data['vdw'] = [ti_vdw_list, ti_vdw_sd_list]

    def compute_convergence_gbar(self, dt=10):
        """
        """

        restr_num = len(self.lambda_restr)
        coul_num = len(self.lambda_coul)

        tot_list = []
        tot_sd_list = []
        coul_list = []
        coul_sd_list = []
        vdw_list = []
        vdw_sd_list = []
        restr_list = []
        restr_sd_list = []

        time_list = np.arange(dt, self.prod_time + dt, dt)
        self.convergence_data = {'time': time_list}

        for time in time_list:
            local_table = FreeEner.get_bar(
                self.xvg_file_list, begin_time=time - dt, end_time=time)
            tot_contrib = local_table['DG (kT)'].sum() * self.conv_fac
            tot_list.append(-tot_contrib)
            tot_std_contrib = sum(self.table['+/-']**2)**0.5 * self.conv_fac
            tot_sd_list.append(tot_std_contrib)

            if self.lambda_restr:
                restr_contrib = local_table['DG (kT)'][:restr_num - 1].sum()
                restr_contrib *= self.conv_fac
                restr_std_contrib = sum(
                    local_table['+/-'][:restr_num - 1]**2)**0.5
                restr_std_contrib *= self.conv_fac
                restr_list.append(-restr_contrib)
                restr_sd_list.append(restr_std_contrib)

            if self.lambda_coul:
                coul_contrib = local_table['DG (kT)'][
                    max(restr_num - 1, 0):restr_num + coul_num - 1].sum()
                coul_contrib *= self.conv_fac
                coul_std_contrib = sum(local_table['+/-'][
                    max(restr_num - 1, 0):restr_num + coul_num - 1]**2)**0.5
                coul_std_contrib *= self.conv_fac
                coul_list.append(-coul_contrib)
                coul_sd_list.append(coul_std_contrib)

            if self.lambda_vdw:
                vdw_contrib = local_table['DG (kT)'][
                    restr_num + coul_num - 1:].sum()
                vdw_contrib *= self.conv_fac
                vdw_std_contrib = sum(
                    local_table['+/-'][restr_num + coul_num - 1:]**2)**0.5
                vdw_std_contrib *= self.conv_fac
                vdw_list.append(-vdw_contrib)
                vdw_sd_list.append(vdw_std_contrib)

        self.convergence_data['tot'] = [tot_list, tot_sd_list]
        if self.lambda_restr:
            self.convergence_data['restr'] = [restr_list, restr_sd_list]
        if self.lambda_coul:
            self.convergence_data['coul'] = [coul_list, coul_sd_list]
        if self.lambda_vdw:
            self.convergence_data['vdw'] = [vdw_list, vdw_sd_list]

    def plot_convergence_graph(self, graph_out=None):
        """
        """
        width_ratio = 0.85

        fig, axs = plt.subplots(2, figsize=(8, 8), sharex=True)
        fig.suptitle(r'Convergence of $\Delta G$')
        fig.subplots_adjust(hspace=0)

        time_list = self.convergence_data['time']

        if self.lambda_restr:
            ti_restr_list = self.convergence_data['restr'][0]
            ti_restr_sd_list = self.convergence_data['restr'][1]
            axs[0].errorbar(
                time_list, ti_restr_list, ti_restr_sd_list,
                label=r'$\Delta G_{restr}$')
        if self.lambda_coul:
            ti_coul_list = self.convergence_data['coul'][0]
            ti_coul_sd_list = self.convergence_data['coul'][1]
            axs[0].errorbar(
                time_list, ti_coul_list, ti_coul_sd_list,
                label=r'$\Delta G_{coul}$')
        if self.lambda_vdw:
            ti_vdw_list = self.convergence_data['vdw'][0]
            ti_vdw_sd_list = self.convergence_data['vdw'][1]
            axs[0].errorbar(
                time_list, ti_vdw_list, ti_vdw_sd_list,
                label=r'$\Delta G_{vdw}$')
        axs[0].legend(bbox_to_anchor=(1.05, 1), loc='upper left')
        axs[0].grid()
        box = axs[0].get_position()
        axs[0].set_position([box.x0, box.y0,
                             box.width * width_ratio,
                             box.height])

        ti_tot_list = self.convergence_data['tot'][0]
        ti_tot_sd_list = self.convergence_data['tot'][1]
        axs[1].errorbar(
            time_list, ti_tot_list, ti_tot_sd_list, label=r'$\Delta G_{tot}$')
        axs[1].legend(bbox_to_anchor=(1.05, 1), loc='upper left')

        axs[1].set_ylabel(r'$\Delta G \; ({})$'.format(self.unit_graph))
        axs[1].yaxis.set_label_coords(-0.1, 1)
        axs[1].set_xlabel(r'$time \; (ps)$')
        axs[1].grid()
        box = axs[1].get_position()
        axs[1].set_position([box.x0, box.y0,
                             box.width * width_ratio,
                             box.height])

        if graph_out:
            plt.savefig(graph_out)

        return(fig)

    @staticmethod
    def symmetry_correction(smile, temp=300):
        r""" Compute symmetry correction
        :math:`\Delta_{sym} = −k T ln(\sigma)`

        return value in kcal/mol

        .. code-block:: python

            > FreeEner.symmetry_correction('c1ccccc1')
            -1.4814...
        """

        try:
            from rdkit import Chem
            # from rdkit.Chem import rdmolfiles
        except ImportError:
            logger.warning('WARNING !!!! \n'
                           'Could not load rdkit \nInstall it using conda:\n'
                           'conda install -c conda-forge rdkit\n'
                           'Symmetry correction set to 0\n')
            return(0.0)

        x = Chem.MolFromSmiles(smile)
        # z = list(rdmolfiles.CanonicalRankAtoms(x, breakTies=False))
        matches = x.GetSubstructMatches(x, uniquify=False)
        sigma = len(matches)
        logger.warning('WARNING !!!! \n'
                       'symmetry_correction() is quite experimental\n'
                       'seems to work for small molecules, use it with'
                       ' caution\n')
        logger.info(f'For molecule {smile} Symmetry number'
                    f' sigma is set to {sigma}')

        GK = - KB * 1e-3 / 4.184  # Gas constant kcal/mol/K

        return GK * temp * math.log(sigma)

    """
    Test with threads:

    def run_threads(self, lambda_coul_list, lambda_vdw_list,
                    lambda_restr_list=[],
                    mbar=False, dir_name='free_ener_run',
                    em_steps=5000, nvt_time=10, npt_time=10, prod_time=100,
                    dt=0.002, temp=300.0, thread_num=4,
                    temp_groups='Protein non-Protein', maxwarn=1,
                    monitor_tool=monitor.PROGRESS_BAR):
        \"""Compute free energy to transfer a molecule from the
        system to vacum.

        :param out_folder: path of the output folder
        :type out_folder: str

        :param mol_name: Name of the molecule
        :type mol_name: str

        :param lambda_coul_list: List lambda points for Coulomb
        :type lambda_coul_list: list

        :param lambda_vdw_list: List lambda points for Lennard Jones
        :type lambda_vdw_list: list

        :param lambda_bond_list: List lambda points for restraints
        :type lambda_bond_list: list, default=[]

        :param mbar: MBAR flag
        :type mbar: bool, default=False

        :param em_steps: number of minimisation steps
        :type em_steps: int, default=5000

        :param nvt_time: Time (ps) of NVT equilibration
        :type nvt_time: int, default=10 ps

        :param npt_time:  Time (ps) of NPT equilibration
        :type npt_time: int, default=10 ps

        :param prod_time: Time (ps) of production run
        :type prod_time: float, default=100 ps

        :param dt: integration time step
        :type dt: float, default=0.002

        :param name: name of the simulation to run
        :type name: str, default=None

        :param temp: Temperature K
        :type temp: float, default=300.0

        :param temp_groups: Group(s) for temperature coupling
        :type temp_groups: str, default='Protein non-Protein'

        :param maxwarn: Maximum number of warnings when using ``gmx grompp``
        :type maxwarn: int, default=0

        :param monitor: option to monitor a simulation, if not none monitor
            should contains two values: ``function`` the function to be ran
            while simulation is running and ``input`` parameters for the
            function
        :type rerun: dict, default=None

        **Object requirement(s):**

            * self.coor_file
            * self.top_file
            * self.nt
            * self.ntmpi
            * self.gpu_id

        **Object field(s) changed:**

            * self.tpr
            * self.sim_name
            * self.coor_file
            * self.xtc

        \"""

        if monitor.isnotebook():
            from tqdm.notebook import tqdm
        else:
            from tqdm import tqdm

        # Remove lambda=0 for vdw and elec
        # If previous restr lambdas are computed
        if len(lambda_restr_list) > 0:
            lambda_coul_list = list(lambda_coul_list)
            lambda_coul_list.remove(0.0)
        if len(lambda_restr_list) + len(lambda_coul_list) > 0:
            lambda_vdw_list = list(lambda_vdw_list)
            lambda_vdw_list.remove(0.0)

        lambda_restr_num = len(lambda_restr_list)
        lambda_coul_num = len(lambda_coul_list)
        lambda_vdw_num = len(lambda_vdw_list)

        restr_lambdas = "".join([
            '{:.2f} '.format(i) for i in lambda_restr_list]) +\
            "1.00 " * (lambda_vdw_num + lambda_coul_num)
        coul_lambdas = "0.00 " * lambda_restr_num + "".join([
            '{:.2f} '.format(i) for i in lambda_coul_list]) +\
            "1.00 " * (lambda_vdw_num)
        vdw_lambdas = "0.00 " * (lambda_coul_num + lambda_restr_num) +\
            "".join(['{:.2f} '.format(i) for i in lambda_vdw_list])

        logger.info(f"Coulomb lambda : {coul_lambdas}\n"
                    f"Vdw lambda     : {vdw_lambdas}\n"
                    f"restr_lambdas  : {restr_lambdas}")

        # https://events.prace-ri.eu/event/674/attachments/618/896/MD_FreeEnergyTutorial.pdf
        # Suggest to use tau_t = 1.0 to avoid
        # over-damping the dynamics of water
        tau_t_str = " ".join(['{}'.format(1.0)
                              for _ in range(len(temp_groups.split()))])
        ref_t_str = " ".join(['{}'.format(temp)
                              for _ in range(len(temp_groups.split()))])

        free_ener_option_md = {'integrator': 'sd',
                               'dt': dt,
                               'constraints': 'all-bonds',
                               'nstcalcenergy': 50,
                               'tcoupl': '',
                               'tc_grps': temp_groups,
                               'tau_t': tau_t_str,
                               'ref_t': ref_t_str,
                               'vdwtype': 'cut_off',
                               'vdw_modifier': 'force_switch',
                               'rvdw_switch': 1.0,
                               'rvdw': 1.1,
                               'coulombtype': 'pme',
                               'rcoulomb': 1.1,
                               'lincs_order': 8,  # Try to fix the LINCS Errors
                               'free_energy': 'yes',
                               'init_lambda-state': 0,
                               'calc-lambda-neighbors': 1,
                               'delta_lambda': 0,
                               'coul_lambdas': coul_lambdas,
                               'vdw_lambdas': vdw_lambdas,
                               'bonded_lambdas': restr_lambdas,
                               'sc_alpha': 0.5,
                               'sc_power': 1,
                               'sc_sigma': 0.3,
                               'couple_moltype': self.mol_name,
                               'couple_lambda0': 'vdw-q',
                               'couple_lambda1': 'none',
                               'couple_intramol': 'no',
                               'nstdhdl': 50,
                               'separate_dhdl_file': 'yes'}

        self.xvg_file_list = []

        nvt_steps = int(nvt_time / dt)
        npt_steps = int(npt_time / dt)
        prod_steps = int(prod_time / dt)

        tot_step = (lambda_restr_num + lambda_coul_num + lambda_vdw_num) * (
            em_steps + nvt_steps + npt_steps + prod_steps)
        pbar = tqdm(total=tot_step)

        from concurrent.futures import ThreadPoolExecutor

        lambda_list = list(
            range(lambda_restr_num + lambda_coul_num + lambda_vdw_num))
        start_dir = os.path.abspath(".")

        def run_lambda(i):
            os.chdir(start_dir)
            print('change dir to:', start_dir)
            logger.info('Compute lambda {} / {}'.format(
                i + 1, lambda_restr_num + lambda_coul_num + lambda_vdw_num))

            lambda_sys = FreeEner.compute_lambda_point(
                self.gmxsys, i,
                self.mol_name,
                os.path.join(self.out_folder,
                             dir_name),
                free_ener_option_md, pbar,
                mbar=mbar,
                em_steps=em_steps,
                nvt_steps=nvt_steps,
                npt_steps=npt_steps,
                prod_steps=prod_steps,
                dt=dt,
                maxwarn=maxwarn,
                monitor_tool=monitor_tool)
            return(lambda_sys)

        with ThreadPoolExecutor(max_workers=thread_num) as executor:
            lambda_results = executor.map(run_lambda, lambda_list)

        for result in lambda_results:
            self.lambda_sys_list.append(result)
            self.xvg_file_list.append(result.xvg)

        self.temp = temp
        self.lambda_coul = list(lambda_coul_list)
        self.lambda_vdw = list(lambda_vdw_list)
        self.lambda_restr = list(lambda_restr_list)
        self.prod_time = prod_time



    def extend_lambda_prod_threads(self, prod_time, thread_num=4):
        \"""
        \"""

        dt = float(self.lambda_sys_list[0].get_mdp_dict()['dt'])
        nsteps = prod_time / dt
        lambda_num = len(self.lambda_restr) + len(self.lambda_coul) +\
            len(self.lambda_vdw)
        xvg_file_list = []
        lambda_list = list(range(lambda_num))
        start_dir = os.path.abspath(".")

        def extend_lambda(lambda_i):
            os.chdir(start_dir)
            logger.info(f'Compute lambda {lambda_i+1} / {lambda_num}')

            self.lambda_sys_list[lambda_i].extend_sim(nsteps=nsteps)
            output = self.lambda_sys_list[lambda_i].get_all_output()
            return(output['xvg'])

        from concurrent.futures import ThreadPoolExecutor
        with ThreadPoolExecutor(max_workers=thread_num) as executor:
            lambda_results = executor.map(extend_lambda, lambda_list)

        for result in lambda_results:
            xvg_file_list += result

        self.prod_time = prod_time
        self.xvg_file_list = xvg_file_list
    """



#!/usr/bin/env python3
# coding: utf-8

# Autorship information
__author__ = "Samuel Murail, Maxence Delaunay, Damien Espana"
__copyright__ = "Copyright 2020, RPBS"
__credits__ = ["Damien Espana", "Maxence Delaunay", "Samuel Murail"]
__license__ = "GNU General Public License v2.0"
__version__ = "2.0.0"
__maintainer__ = "Samuel Murail"
__email__ = "samuel.murail@u-paris.fr"
__status__ = "Prototype"
